import re
from unittest import mock

import pytest
import responses
import yaml

from cumulusci.core.exceptions import BulkDataException
from cumulusci.tasks.bulkdata import LoadData
from cumulusci.tasks.bulkdata.step import DataApi, DataOperationStatus
from cumulusci.tests.util import CURRENT_SF_API_VERSION, mock_describe_calls


class TestUpsert:
    # Would be nice to slim down this VCR further. It's still 50kb.
    # The next step of VCR compression would be to have some templates
    # for XML that can be reused when matched.
    # gzip would be another (albeit binary) answer.
    @pytest.mark.vcr()
    def test_upsert_external_id_field(
        self,
        create_task,
        cumulusci_test_repo_root,
        sf,
        delete_data_from_org,
        run_code_without_recording,
    ):
        run_code_without_recording(
            lambda: delete_data_from_org(
                ["Entitlement", "Opportunity", "Contact", "Account"]
            )
        )
        self._test_two_upserts_and_check_results(
            "bulk", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr()
    def test_simple_upsert__rest(
        self,
        create_task,
        cumulusci_test_repo_root,
        run_code_without_recording,
        delete_data_from_org,
        sf,
    ):

        run_code_without_recording(
            lambda: delete_data_from_org(
                ["Entitlement", "Opportunity", "Contact", "Account"]
            )
        )
        self._test_two_upserts_and_check_results(
            "rest",
            create_task,
            cumulusci_test_repo_root,
            sf,
        )

    def _test_two_upserts_and_check_results(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        # seed sample data, using a mixture of inserts and
        # upserts-into-empty (which should behave as inserts)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_before_data.sql",
                "mapping": cumulusci_test_repo_root
                / f"datasets/upsert/upsert_mapping_{api}.yml",
                # "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )

        # load the data and double-check that it loaded as expected
        with mock.patch.object(task.logger, "info"):
            task()
            assert "upsert" in str(task.logger.info.mock_calls)

        # check that the right steps were executed
        result = task.return_values
        assert result == {
            "step_results": {
                "Insert Accounts": {
                    "sobject": "Account",
                    "record_type": None,
                    "status": DataOperationStatus.SUCCESS,
                    "job_errors": [],
                    "records_processed": 1,
                    "total_row_errors": 0,
                },
                "Upsert Contacts": {
                    "sobject": "Contact",
                    "record_type": None,
                    "status": DataOperationStatus.SUCCESS,
                    "job_errors": [],
                    "records_processed": 16,
                    "total_row_errors": 0,
                },
                "Insert Opportunities": {
                    "sobject": "Opportunity",
                    "record_type": None,
                    "status": DataOperationStatus.SUCCESS,
                    "job_errors": [],
                    "records_processed": 0,
                    "total_row_errors": 0,
                },
            }
        }, result

        # check that Salesforce contains what we expectdd
        accounts = sf.query("select Name from Account")
        accounts = {account["Name"] for account in accounts["records"]}
        assert "Sitwell-Bluth" in accounts
        contacts = sf.query("select FirstName from Contact")["records"]
        firstnames = {contact["FirstName"] for contact in contacts}
        assert "Nichael" not in firstnames
        assert "George Oscar" not in firstnames

        # Upsert against the data we already created
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / f"datasets/upsert/upsert_mapping_{api}.yml",
                # "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task()

        # check that the right data was loaded
        contacts = sf.query(
            "select FirstName,(select Name from Opportunities) from Contact"
        )["records"]
        firstnames = {contact["FirstName"] for contact in contacts}

        assert "Nichael" in firstnames
        assert "George Oscar" in firstnames
        opportunity_names = [
            contact["Opportunities"]["records"][0]["Name"]
            for contact in contacts
            if contact["Opportunities"]
        ]
        assert set(opportunity_names) == set(
            ["Espionage Opportunity", "Illusional Opportunity"]
        ), set(opportunity_names)

    @responses.activate
    def test_upsert_rest__faked(
        self, create_task, cumulusci_test_repo_root, org_config
    ):
        domain = org_config.get_domain()
        ver = CURRENT_SF_API_VERSION
        responses.add(
            method="PATCH",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects/Contact/Email",
            status=200,
            json=[
                {
                    "id": "003P000001Y5exdIAB",
                    "success": True,
                    "errors": [],
                    "created": True,
                },
                {
                    "id": "003P000001Y5exXIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
                {
                    "id": "003P000001Y5exYIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
            ],
        )
        responses.add(
            method="POST",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects",
            status=200,
            json=[{"id": "001J000002jcjvnIAA", "success": True, "errors": []}],
        )
        mock_describe_calls(domain=domain, version=CURRENT_SF_API_VERSION)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_mapping_rest.yml",
                "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task._update_credentials = mock.Mock()
        with mock.patch.object(task.logger, "debug"):
            rc = task()
            assert rc == {
                "step_results": {
                    "Insert Accounts": {
                        "sobject": "Account",
                        "record_type": None,
                        "status": DataOperationStatus.SUCCESS,
                        "job_errors": [],
                        "records_processed": 0,
                        "total_row_errors": 0,
                    },
                    "Upsert Contacts": {
                        "sobject": "Contact",
                        "record_type": None,
                        "status": DataOperationStatus.SUCCESS,
                        "job_errors": [],
                        "records_processed": 3,
                        "total_row_errors": 0,
                    },
                    "Insert Opportunities": {
                        "sobject": "Opportunity",
                        "record_type": None,
                        "status": DataOperationStatus.SUCCESS,
                        "job_errors": [],
                        "records_processed": 1,
                        "total_row_errors": 0,
                    },
                }
            }, rc

            relevant_debug_statement = look_for_operation_creation_debug_statement(
                task.logger.debug.mock_calls
            )
            assert relevant_debug_statement == format(
                DataApi.REST
            ), relevant_debug_statement

    def _mock_bulk(self, domain):
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{CURRENT_SF_API_VERSION}/limits/recordCount?sObjects=Contact",
            status=200,
            json={"sObjects": []},
        )

        responses.add(
            method="POST",
            url=f"https://{domain}/services/async/{CURRENT_SF_API_VERSION}/job",
            status=200,
            body=f"""<?xml version="1.0" encoding="UTF-8"?>
            <jobInfo xmlns="http://www.force.com/2009/06/asyncapi/dataload">
        <id>750P0000005mX3LIAU</id>
        <operation>insert</operation>
        <object>Account</object>
        <createdById>005P0000009ajX7IAI</createdById>
        <createdDate>2022-03-02T04:14:43.000Z</createdDate>
        <systemModstamp>2022-03-02T04:14:43.000Z</systemModstamp>
        <state>Open</state>
        <concurrencyMode>Parallel</concurrencyMode>
        <contentType>CSV</contentType>
        <numberBatchesQueued>0</numberBatchesQueued>
        <numberBatchesInProgress>0</numberBatchesInProgress>
        <numberBatchesCompleted>0</numberBatchesCompleted>
        <numberBatchesFailed>0</numberBatchesFailed>
        <numberBatchesTotal>0</numberBatchesTotal>
        <numberRecordsProcessed>0</numberRecordsProcessed>
        <numberRetries>0</numberRetries>
        <apiVersion>v{CURRENT_SF_API_VERSION}</apiVersion>
        <numberRecordsFailed>0</numberRecordsFailed>
        <totalProcessingTime>0</totalProcessingTime>
        <apiActiveProcessingTime>0</apiActiveProcessingTime>
        <apexProcessingTime>0</apexProcessingTime>
        </jobInfo>""",
        )
        responses.add(
            method="POST",
            url=f"https://{domain}/services/async/{CURRENT_SF_API_VERSION}/job/750P0000005mX3LIAU",
            status=200,
            body=f""""<?xml version="1.0" encoding="UTF-8"?>
            <jobInfo xmlns="http://www.force.com/2009/06/asyncapi/dataload">
            <id>750P0000005mX3LIAU</id>
            <operation>insert</operation>
            <object>Account</object>
            <createdById>005P0000009ajX7IAI</createdById>
            <createdDate>2022-03-02T04:14:43.000Z</createdDate>
            <systemModstamp>2022-03-02T04:14:43.000Z</systemModstamp>
            <state>Closed</state>
            <concurrencyMode>Parallel</concurrencyMode>
            <contentType>CSV</contentType>
            <numberBatchesQueued>0</numberBatchesQueued>
            <numberBatchesInProgress>0</numberBatchesInProgress>
            <numberBatchesCompleted>0</numberBatchesCompleted>
            <numberBatchesFailed>0</numberBatchesFailed>
            <numberBatchesTotal>0</numberBatchesTotal>
            <numberRecordsProcessed>0</numberRecordsProcessed>
            <numberRetries>0</numberRetries>
            <apiVersion>v{CURRENT_SF_API_VERSION}.0</apiVersion>
            <numberRecordsFailed>0</numberRecordsFailed>
            <totalProcessingTime>0</totalProcessingTime>
            <apiActiveProcessingTime>0</apiActiveProcessingTime>
            <apexProcessingTime>0</apexProcessingTime>
            </jobInfo>""",
        )

        responses.add(
            method="GET",
            url=f"https://{domain}/services/async/{CURRENT_SF_API_VERSION}/job/750P0000005mX3LIAU",
            status=200,
            body="""<?xml version="1.0" encoding="UTF-8"?>
            <jobInfo xmlns="http://www.force.com/2009/06/asyncapi/dataload">
            <id>750P0000005mX3LIAU</id>
            <operation>insert</operation>
            <object>Account</object>
            <createdById>005P0000009ajX7IAI</createdById>
            <createdDate>2022-03-02T04:14:43.000Z</createdDate>
            <systemModstamp>2022-03-02T04:14:43.000Z</systemModstamp>
            <state>Closed</state>
            <concurrencyMode>Parallel</concurrencyMode>
            <contentType>CSV</contentType>
            <numberBatchesQueued>0</numberBatchesQueued>
            <numberBatchesInProgress>0</numberBatchesInProgress>
            <numberBatchesCompleted>0</numberBatchesCompleted>
            <numberBatchesFailed>0</numberBatchesFailed>
            <numberBatchesTotal>0</numberBatchesTotal>
            <numberRecordsProcessed>4</numberRecordsProcessed>
            <numberRetries>0</numberRetries>
            <apiVersion>vxx.0</apiVersion>
            <numberRecordsFailed>0</numberRecordsFailed>
            <totalProcessingTime>0</totalProcessingTime>
            <apiActiveProcessingTime>0</apiActiveProcessingTime>
            <apexProcessingTime>0</apexProcessingTime>
            </jobInfo>""",
        )
        responses.add(
            method="GET",
            url=f"https://{domain}/services/async/{CURRENT_SF_API_VERSION}/job/750P0000005mX3LIAU/batch",
            status=200,
            body="""<?xml version="1.0" encoding="UTF-8"?>
            <batchInfoList xmlns="http://www.force.com/2009/06/asyncapi/dataload"/>""",
        )

        responses.add(
            method="POST",
            url=f"https://{domain}/services/async/{CURRENT_SF_API_VERSION}/job/750P0000005mX3LIAU/batch",
            status=200,
            body="""<?xml version="1.0" encoding="UTF-8"?>
            <batchInfo xmlns="http://www.force.com/2009/06/asyncapi/dataload">
        <id>751P0000006fIUAIA2</id>
        <jobId>750P0000005mX2wIAE</jobId>
        <state>Queued</state>
        <createdDate>2022-03-02T04:14:44.000Z</createdDate>
        <systemModstamp>2022-03-02T04:14:44.000Z</systemModstamp>
        <numberRecordsProcessed>4</numberRecordsProcessed>
        <numberRecordsFailed>0</numberRecordsFailed>
        <totalProcessingTime>0</totalProcessingTime>
        <apiActiveProcessingTime>0</apiActiveProcessingTime>
        <apexProcessingTime>0</apexProcessingTime>
        </batchInfo>""",
        )

        responses.add(
            method="GET",
            url=f"https://{domain}/services/async/{CURRENT_SF_API_VERSION}/job/750P0000005mX3LIAU/batch/751P0000006fIUAIA2/result",
            status=200,
            body="""""Id","Success","Created","Error"
        "003P000001Y5yvdIAB","True","false",""
        "003P000001Y5ywMIAR","True","false",""
        "003P000001Y5ywNIAR","True","false",""
        "003P000001Y5ywOIAR","True","false","" """,
        )

    @responses.activate
    def test_upsert__fake_bulk(self, create_task, cumulusci_test_repo_root, org_config):
        domain = org_config.get_domain()
        mock_describe_calls(domain=domain, version=CURRENT_SF_API_VERSION)
        self._mock_bulk(domain)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_mapping_bulk.yml",
                "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task._update_credentials = mock.Mock()

        with mock.patch.object(task.logger, "debug"):
            ret = task()
            assert ret == {
                "step_results": {
                    "Insert Accounts": {
                        "sobject": "Account",
                        "record_type": None,
                        "status": DataOperationStatus.SUCCESS,
                        "job_errors": [],
                        "records_processed": 0,
                        "total_row_errors": 0,
                    },
                    "Upsert Contacts": {
                        "sobject": "Contact",
                        "record_type": None,
                        "status": DataOperationStatus.SUCCESS,
                        "job_errors": [],
                        "records_processed": 0,  # change here and above to 4 to match data
                        "total_row_errors": 0,
                    },
                    "Insert Opportunities": {
                        "sobject": "Opportunity",
                        "record_type": None,
                        "status": DataOperationStatus.SUCCESS,
                        "job_errors": [],
                        "records_processed": 0,
                        "total_row_errors": 0,
                    },
                }
            }, ret
            relevant_debug_statement = look_for_operation_creation_debug_statement(
                task.logger.debug.mock_calls
            )
            assert relevant_debug_statement in format(
                DataApi.BULK
            ), relevant_debug_statement

    def _test_two_upserts_and_check_results__complex(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_before_data.sql",
                "mapping": cumulusci_test_repo_root
                / f"datasets/upsert/upsert_mapping_{api}_complex.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task()
        result = task.return_values
        assert all(
            val["status"] == DataOperationStatus.SUCCESS
            for val in result["step_results"].values()
        ), result.values()
        contacts = sf.query("select Email from Contact")
        emails = {contacts["Email"] for contacts in contacts["records"]}
        assert "michael.bluth@example.com" in emails, emails
        assert "nichael.bluth@example.com" not in emails, emails
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_complex_example.sql",
                "mapping": cumulusci_test_repo_root
                / f"datasets/upsert/upsert_mapping_{api}_complex.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task()
        contacts = sf.query("select Email from Contact")
        emails = {contacts["Email"] for contacts in contacts["records"]}
        assert "michael.bluth@example.com" not in emails
        assert "nichael.bluth@example.com" in emails

    @pytest.mark.vcr()
    def test_upsert_complex_fields__bulk(
        self,
        create_task,
        cumulusci_test_repo_root,
        sf,
        delete_data_from_org,
        run_code_without_recording,
        # mock_bulk_download_for_vcr,
    ):
        run_code_without_recording(
            lambda: delete_data_from_org(
                ["Entitlement", "Opportunity", "Contact", "Account"]
            )
        )
        self._test_two_upserts_and_check_results__complex(
            "bulk", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr()
    def test_upsert_complex_external_id_field__rest(
        self,
        create_task,
        cumulusci_test_repo_root,
        sf,
        run_code_without_recording,
        delete_data_from_org,
    ):
        run_code_without_recording(
            lambda: delete_data_from_org(
                ["Entitlement", "Opportunity", "Contact", "Account"]
            )
        )
        self._test_two_upserts_and_check_results__complex(
            "rest", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr()
    def test_upsert_complex_external_id_field_rest__duplicate_error(
        self,
        create_task,
        cumulusci_test_repo_root,
        sf,
        run_code_without_recording,
        delete_data_from_org,
    ):
        """Test what happens if the upsert key is duplicate and therefore it is unclear
        which record to update."""

        run_code_without_recording(
            lambda: delete_data_from_org(
                ["Entitlement", "Opportunity", "Contact", "Account"]
            )
        )

        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_before_data.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_mapping_rest_complex.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task()

        # inject a duplicate to cause an error
        assert sf.query("select Name from Account where AccountNumber='420'")[
            "records"
        ][0]["Name"]
        assert sf.Account.create(
            {"Name": "xyzzy", "AccountNumber": "420"}
        )  # make a dupe

        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_complex_example.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_mapping_rest_complex.yml",
                "set_recently_viewed": False,
            },
        )
        with pytest.raises(BulkDataException) as e:
            task()
        assert "Duplicate values for upsert key:" in str(e.value)

    @responses.activate
    def test_simple_upsert__bad_key(
        self, create_task, cumulusci_test_repo_root, org_config
    ):
        domain = org_config.get_domain()
        ver = CURRENT_SF_API_VERSION
        responses.add(
            method="PATCH",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects/Contact/Email",
            status=200,
            json=[
                {
                    "id": "003P000001Y5exdIAB",
                    "success": True,
                    "errors": [],
                    "created": True,
                },
                {
                    "id": "003P000001Y5exXIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
                {
                    "id": "003P000001Y5exYIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
            ],
        )
        responses.add(
            method="POST",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects",
            status=200,
            json=[{"id": "001J000002jcjvnIAA", "success": True, "errors": []}],
        )
        mock_describe_calls(domain=domain, version=CURRENT_SF_API_VERSION)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_mapping_simple__bad_key.yml",
                "ignore_row_errors": True,
                "set_recently_viewed": False,
            },
        )
        with pytest.raises(BulkDataException) as e:
            task._update_credentials = mock.Mock()
            task()
        assert "FirstName" in str(e.value)
        assert "update key" in str(e.value)

    @responses.activate
    def test_simple_upsert__smart(
        self, create_task, cumulusci_test_repo_root, org_config
    ):
        domain = org_config.get_domain()
        ver = CURRENT_SF_API_VERSION
        responses.add(
            method="PATCH",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects/Contact/Email",
            status=200,
            json=[
                {
                    "id": "003P000001Y5exdIAB",
                    "success": True,
                    "errors": [],
                    "created": True,
                },
                {
                    "id": "003P000001Y5exXIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
                {
                    "id": "003P000001Y5exYIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
            ],
        )
        responses.add(
            method="POST",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects",
            status=200,
            json=[{"id": "001J000002jcjvnIAA", "success": True, "errors": []}],
        )
        mock_describe_calls(domain=domain, version=CURRENT_SF_API_VERSION)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_smart__native_field.yml",
                "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task._update_credentials = mock.Mock()
        task()
        assert task.return_values == {
            "step_results": {
                "Upsert Contacts": {
                    "sobject": "Contact",
                    "record_type": None,
                    "status": DataOperationStatus.SUCCESS,
                    "job_errors": [],
                    "records_processed": 3,
                    "total_row_errors": 0,
                }
            }
        }

    @responses.activate
    def test_simple_upsert_smart__native_field(
        self, create_task, cumulusci_test_repo_root, org_config
    ):
        domain = org_config.get_domain()
        ver = CURRENT_SF_API_VERSION

        # do a native UPSERT
        responses.add(
            method="PATCH",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects/Contact/Email",
            status=200,
            json=[
                {
                    "id": "003P000001Y5exdIAB",
                    "success": True,
                    "errors": [],
                    "created": True,
                },
                {
                    "id": "003P000001Y5exXIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
                {
                    "id": "003P000001Y5exYIAR",
                    "success": True,
                    "errors": [],
                    "created": False,
                },
            ],
        )
        responses.add(
            method="POST",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects",
            status=200,
            json=[{"id": "001J000002jcjvnIAA", "success": True, "errors": []}],
        )
        mock_describe_calls(domain=domain, version=CURRENT_SF_API_VERSION)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_smart__native_field.yml",
                "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task._update_credentials = mock.Mock()
        task()
        assert task.return_values == {
            "step_results": {
                "Upsert Contacts": {
                    "sobject": "Contact",
                    "record_type": None,
                    "status": DataOperationStatus.SUCCESS,
                    "job_errors": [],
                    "records_processed": 3,
                    "total_row_errors": 0,
                }
            }
        }

    @responses.activate
    def test_upsert_recordtype_same_developername_different_ispersontype(
        self, create_task, cumulusci_test_repo_root, org_config, sf
    ):
        domain = org_config.get_domain()
        ver = CURRENT_SF_API_VERSION
        expected_number_of_records = 3
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/query/?q=SELECT+Id%2C+DeveloperName%2C+IsPersonType+FROM+RecordType+WHERE+SObjectType%3D%27Account%27",
            status=200,
            json={
                "totalSize": 4,
                "done": True,
                "records": [
                    {
                        "Id": "0125j000000RqVkAAK",
                        "DeveloperName": "HH_Account",
                        "IsPersonType": False,
                    },
                    {
                        "Id": "0125j000000RqVlAAK",
                        "DeveloperName": "Organization",
                        "IsPersonType": False,
                    },
                    {
                        "Id": "0125j000000bo4yAAA",
                        "DeveloperName": "PersonAccount",
                        "IsPersonType": True,
                    },
                    {
                        "Id": "0125j000000bo53AAA",
                        "DeveloperName": "PersonAccount",
                        "IsPersonType": False,
                    },
                ],
            },
        )
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/limits/recordCount?sObjects=Account",
            status=200,
            json={"sObjects": [{"count": 3, "name": "Account"}]},
        )
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/query/?q=select+Id%2CAccountNumber+from+Account",
            status=200,
            json={
                "totalSize": 3,
                "done": True,
                "records": [
                    {"Id": "0015j00001H0q4NAAR", "AccountNumber": "12345"},
                    {"Id": "0015j00001H0q4OAAR", "AccountNumber": "456789"},
                    {"Id": "0015j00001H0q7bAAB", "AccountNumber": "909098"},
                ],
            },
        )
        with (
            cumulusci_test_repo_root
            / "cumulusci/tests/cassettes/GET_sobjects_Account_PersonAccount_describe.yaml"
        ).open("r") as f:
            body_accounts = yaml.safe_load(f)["response"]["body"]["string"]
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/sobjects/Account/describe",
            body=body_accounts,
            status=200,
        )
        with (
            cumulusci_test_repo_root
            / "cumulusci/tests/shared_cassettes/GET_sobjects_Global_describe.yaml"
        ).open("r") as f:
            body_global = yaml.safe_load(f)["response"]["body"]["string"]
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/sobjects",
            body=body_global,
            status=200,
        )
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_recordtypes.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_mapping_recordtypes.yml",
                "set_recently_viewed": False,
            },
        )
        task._update_credentials = mock.Mock()
        task.sf = sf
        task.bulk = mock.Mock()
        task._init_mapping()
        with task._init_db():
            task._expand_mapping()
            mapping = task.mapping["Upsert Accounts"]
            if "RecordTypeId" in mapping.fields:
                conn = task.session.connection()
                task._load_record_types([mapping.sf_object], conn)
                task.session.commit()
            _, query = task.configure_step(mapping)
            # Assert no duplicate records are trying to be deployed
            assert len(list(query)) == expected_number_of_records

    @responses.activate
    def test_simple_upsert_smart__non_native_field(
        self, create_task, cumulusci_test_repo_root, org_config
    ):
        domain = org_config.get_domain()
        ver = CURRENT_SF_API_VERSION
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/limits/recordCount?sObjects=Contact",
            status=200,
            json={"sObjects": []},
        )
        # It needs to do ETL, so this should be hit
        responses.add(
            method="GET",
            url=f"https://{domain}/services/data/v{ver}/query/?q=select+Id%2CFirstName+from+Contact",
            status=200,
            json={"totalSize": 0, "done": True, "records": []},
        )
        responses.add(
            method="PATCH",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects/Contact/Id",
            status=200,
            json=[
                {
                    "id": "0035500001GzhXkAAJ",
                    "success": True,
                    "errors": [],
                    "created": False,
                }
            ],
        )

        responses.add(
            method="POST",
            url=f"https://{domain}/services/data/v{ver}/composite/sobjects",
            status=200,
            json=[{"id": "001J000002jcjvnIAA", "success": True, "errors": []}],
        )
        mock_describe_calls(domain=domain, version=CURRENT_SF_API_VERSION)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/upsert/upsert_example_2.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/upsert/upsert_smart__non_native_field.yml",
                "ignore_row_errors": True,
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )
        task._update_credentials = mock.Mock()
        task()
        assert task.return_values == {
            "step_results": {
                "Upsert Contacts": {
                    "sobject": "Contact",
                    "record_type": None,
                    "status": DataOperationStatus.SUCCESS,
                    "job_errors": [],
                    "records_processed": 1,
                    "total_row_errors": 0,
                }
            }
        }


def look_for_operation_creation_debug_statement(mock_calls):
    relevant_debug_statements = (
        look_for_operation_creation_debug_statement_for_string(call.args[0])
        for call in mock_calls
    )

    return next(stmt for stmt in relevant_debug_statements if stmt)


DEBUG_MATCHER = re.compile("Creating (.*) Operation for (.*) using (.*)")


def look_for_operation_creation_debug_statement_for_string(s):
    match = DEBUG_MATCHER.match(s)
    if match:
        return match[3]



import collections
import logging
import tempfile
import typing as T
from contextlib import contextmanager, nullcontext
from pathlib import Path

from simple_salesforce import Salesforce
from sqlalchemy import Boolean, Column, MetaData, Table, Unicode, inspect
from sqlalchemy.engine.base import Connection
from sqlalchemy.orm import Session, mapper

from cumulusci.core.exceptions import BulkDataException
from cumulusci.utils.iterators import iterate_in_chunks


class SqlAlchemyMixin:
    logger: logging.Logger
    metadata: MetaData
    models: dict
    options: dict
    session: Session
    sf: Salesforce

    def _create_record_type_table(self, table_name):
        """Create a table to store mapping between Record Type Ids and Developer Names."""
        rt_map_model_name = f"{table_name}Model"
        self.models[table_name] = type(rt_map_model_name, (object,), {})
        rt_map_fields = [
            Column("record_type_id", Unicode(18), primary_key=True),
            Column("developer_name", Unicode(255)),
            Column("is_person_type", Boolean),
        ]
        rt_map_table = Table(table_name, self.metadata, *rt_map_fields)
        mapper(self.models[table_name], rt_map_table)

    def _extract_record_types(
        self, sobject, tablename: str, conn, is_person_accounts_enabled: bool
    ):
        """Query for Record Type information and persist it in the database."""
        self.logger.info(f"Extracting Record Types for {sobject}")
        if is_person_accounts_enabled:
            query = f"SELECT Id, DeveloperName, IsPersonType FROM RecordType WHERE SObjectType='{sobject}'"
        else:
            query = f"SELECT Id, DeveloperName FROM RecordType WHERE SObjectType='{sobject}'"

        result = self.sf.query(query)

        if result["totalSize"]:
            sql_bulk_insert_from_records(
                connection=conn,
                table=self.metadata.tables[tablename],
                columns=["record_type_id", "developer_name", "is_person_type"],
                record_iterable=(
                    [rt["Id"], rt["DeveloperName"], rt.get("IsPersonType", False)]
                    for rt in result["records"]
                ),
            )

    @contextmanager
    def _temp_database_url(self):
        with tempfile.TemporaryDirectory() as t:
            tempdb = Path(t) / "temp_db.db"

            self.logger.debug(f"Using temporary database {tempdb}")
            database_url = f"sqlite:///{tempdb}"
            yield database_url

    def _database_url(self):
        database_url = self.options.get("database_url")
        if database_url:
            return nullcontext(enter_result=database_url)
        else:
            return self._temp_database_url()


def _handle_primary_key(mapping, fields):
    """Provide support for legacy mappings which used the OID as the pk but
    default to using an autoincrementing int pk and a separate sf_id column"""

    id_column = "id"
    if mapping.get_oid_as_pk():
        # Get Id column from mapping
        id_column = mapping.fields["Id"]

    fields.append(Column(id_column, Unicode(255), primary_key=True))


def create_table(mapping, metadata) -> Table:
    """Given a mapping data structure (from mapping.yml) and SQLAlchemy
    metadata, create a table matching the mapping.

    Mapping should be a MappingStep instance"""

    fields = []
    _handle_primary_key(mapping, fields)

    # make a field list to create
    for field, db in mapping.get_complete_field_map().items():
        if field == "Id":
            continue

        fields.append(Column(db, Unicode(255)))

    if mapping.record_type:
        fields.append(Column("record_type", Unicode(255)))
    return create_table_if_needed(mapping.table, metadata, fields)


def create_table_if_needed(tablename, metadata, fields: T.List[Column]) -> Table:
    t = Table(tablename, metadata, *fields)
    inspector = inspect(metadata.bind)
    if inspector.has_table(tablename):
        raise BulkDataException(f"Table already exists: {tablename}")
    t.create(metadata.bind)
    return t


def generate_batches(num_records, batch_size):
    """Generate batch size list for splitting a number of tasks into batch jobs.

    Given a number of records to split up, and a batch size, generate a
    stream of batchsize, index pairs"""
    num_batches, left_over = divmod(num_records, batch_size)
    if left_over:
        num_batches += 1  # need an extra batch for the clean-up
    for i in range(0, num_batches):
        is_last_batch = i == num_batches - 1
        if is_last_batch and left_over:
            batch_size = left_over
        if batch_size > 0:
            yield batch_size, i, num_batches


class RowErrorChecker:
    def __init__(self, logger, ignore_row_errors, row_warning_limit):
        self.logger = logger
        self.ignore_row_errors = ignore_row_errors
        self.row_warning_limit = row_warning_limit
        self.row_error_count = 0

    def check_for_row_error(self, result, row_id):
        if not result.success:
            msg = f"Error on record with id {row_id}: {result.error}"
            if self.ignore_row_errors:
                if self.row_error_count < self.row_warning_limit:
                    self.logger.warning(msg)
                elif self.row_error_count == self.row_warning_limit:
                    self.logger.warning("Further warnings suppressed")
                self.row_error_count += 1
                return self.row_error_count
            else:
                raise BulkDataException(msg)


def consume(iterator):
    """Consume an iterator for its side effects.

    Simplified from the function in https://docs.python.org/3/library/itertools.html
    """
    collections.deque(iterator, maxlen=0)


def sql_bulk_insert_from_records(
    *,
    connection: Connection,
    table: Table,
    columns: T.Tuple[str],
    record_iterable: T.Iterable,
) -> None:
    """Persist records from the given generator into the local database."""
    consume(
        sql_bulk_insert_from_records_incremental(
            connection=connection,
            table=table,
            columns=columns,
            record_iterable=record_iterable,
        )
    )


def sql_bulk_insert_from_records_incremental(
    *,
    connection: Connection,
    table: Table,
    columns: T.Tuple[str],
    record_iterable: T.Iterable,
):
    """Generator that persists batches of records from the given generator into the local database

    Yields after every batch."""
    dict_iterable = (dict(zip(columns, row)) for row in record_iterable)
    for group in iterate_in_chunks(10000, dict_iterable):
        with connection.begin():
            connection.execute(table.insert(), group)
        # self.session.flush()  -- Did this line do anything?
        yield


def sf_query_to_table(
    *,
    table: Table,
    metadata: MetaData,
    connection: Connection,
    sobject: str,
    fields: T.List[str],
    **queryargs,
) -> Table:
    """Cache data from Salesforce in a SQL Alchemy Table."""
    from cumulusci.tasks.bulkdata.step import DataOperationStatus, get_query_operation

    qs = get_query_operation(sobject=sobject, fields=fields, **queryargs)
    qs.query()
    if qs.job_result.status is not DataOperationStatus.SUCCESS:  # pragma: no cover
        raise BulkDataException(
            f"Unable to query records for {sobject}: {','.join(qs.job_result.job_errors)}"
        )
    results = qs.get_results()

    sql_bulk_insert_from_records(
        connection=connection,
        table=table,
        columns=tuple(fields),
        record_iterable=results,
    )
    return table



from warnings import warn


class ClassMovedWarning(FutureWarning):
    pass


def warn_moved(new, old):
    msg = f"Deprecation warning: Please import from {new} rather than {old}"

    warn(msg, ClassMovedWarning)



# for backwards compatibility
from cumulusci.tasks.preflight.packages import *  # noqa
from cumulusci.utils.deprecation import warn_moved

warn_moved(
    "cumulusci.tasks.preflight.packages",
    __name__,
)



from .github import GitHubSource
from .local_folder import LocalFolderSource


class NullSource:
    """Source for the root project."""

    frozenspec = None
    allow_remote_code = True
    location = None

    def __str__(self):
        return "current folder"


__all__ = ("GitHubSource", "LocalFolderSource", "NullSource")



import pytest

from cumulusci.utils.git import (
    EMPTY_URL_MESSAGE,
    construct_release_branch_name,
    get_release_identifier,
    is_release_branch,
    is_release_branch_or_child,
    parse_repo_url,
    split_repo_url,
)


def test_is_release_branch():
    assert is_release_branch("feature/230", "feature/")
    assert not is_release_branch("feature/230__test", "feature/")
    assert not is_release_branch("main", "feature/")


def test_is_release_branch_or_child():
    assert is_release_branch_or_child("feature/230", "feature/")
    assert is_release_branch_or_child("feature/230__test", "feature/")
    assert is_release_branch_or_child("feature/230__test__gc", "feature/")
    assert not is_release_branch_or_child("main", "feature/")


def test_get_release_identifier():
    assert get_release_identifier("feature/230", "feature/") == "230"
    assert get_release_identifier("feature/230__test", "feature/") == "230"
    assert get_release_identifier("main", "feature/") is None


def test_construct_release_branch_name():
    assert construct_release_branch_name("feature/", "230") == "feature/230"


@pytest.mark.parametrize(
    "repo_uri,owner,repo_name,host",
    [
        (
            "https://git.ent.example.com/org/private_repo/",
            "org",
            "private_repo",
            "git.ent.example.com",
        ),
        ("https://github.com/owner/repo_name/", "owner", "repo_name", "github.com"),
        ("https://github.com/owner/repo_name.git", "owner", "repo_name", "github.com"),
        (
            "https://user@github.com/owner/repo_name.git",
            "owner",
            "repo_name",
            "github.com",
        ),
        (
            "https://git.ent.example.com/org/private_repo.git",
            "org",
            "private_repo",
            "git.ent.example.com",
        ),
        ("git@github.com:owner/repo_name.git", "owner", "repo_name", "github.com"),
        ("git@github.com:/owner/repo_name.git", "owner", "repo_name", "github.com"),
        ("git@github.com:owner/repo_name", "owner", "repo_name", "github.com"),
        (
            "git@api.github.com/owner/repo_name/",
            "owner",
            "repo_name",
            "api.github.com",
        ),
        (
            "git@api.github.com/owner/repo_name.git",
            "owner",
            "repo_name",
            "api.github.com",
        ),
    ],
)
def test_parse_repo_url(repo_uri, owner, repo_name, host):
    assert parse_repo_url(repo_uri) == (owner, repo_name, host)
    assert split_repo_url(repo_uri) == (owner, repo_name)


@pytest.mark.parametrize("URL", [None, ""])
def test_empty_url(URL):
    with pytest.raises(ValueError, match=EMPTY_URL_MESSAGE):
        parse_repo_url(URL)



import responses

from cumulusci.tasks.salesforce.create_permission_sets import CreatePermissionSet
from cumulusci.tests.util import CURRENT_SF_API_VERSION

from .util import create_task


class TestCreatePermissionSet:
    @responses.activate
    def test_create_permset(self):
        task = create_task(
            CreatePermissionSet,
            {
                "api_name": "PermSet",
                "label": "Permission Set",
                "user_permissions": [
                    "PermissionsBulkApiHardDelete",
                    "PermissionsCreateAuditFields",
                ],
            },
        )

        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/sobjects/PermissionSet/",
            status=200,
            json={"id": "0PS3F000000fCNPWA2", "success": True, "errors": []},
        )
        responses.add(
            method="POST",
            url=f"{task.org_config.instance_url}/services/data/v{CURRENT_SF_API_VERSION}/sobjects/PermissionSetAssignment/",
            status=200,
            json={"id": "0Pa000000000001", "success": True, "errors": []},
        )

        task()

        assert len(responses.calls) == 2
        assert "PermissionsBulkApiHardDelete" in responses.calls[0].request.body
        assert "PermissionsCreateAuditFields" in responses.calls[0].request.body



import base64
import io
import os
import pathlib
import zipfile

import pytest

from cumulusci.salesforce_api.package_zip import (
    BasePackageZipBuilder,
    CreatePackageZipBuilder,
    DestructiveChangesZipBuilder,
    InstallPackageZipBuilder,
    MetadataPackageZipBuilder,
    UninstallPackageZipBuilder,
)
from cumulusci.utils import temporary_dir, touch


class TestBasePackageZipBuilder:
    def test_as_hash(self):
        builder = BasePackageZipBuilder()
        builder.zf.writestr("1", "1")
        hash1 = builder.as_hash()

        builder = BasePackageZipBuilder()
        builder.zf.writestr("1", "1")
        hash2 = builder.as_hash()

        assert hash2 == hash1

        builder.zf.writestr("2", "2")
        hash3 = builder.as_hash()
        assert hash3 != hash2


class TestMetadataPackageZipBuilder:
    def test_builder(self, task_context):
        with temporary_dir() as path:

            # add package.xml
            with open(os.path.join(path, "package.xml"), "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <version>45.0</version>
</Package>"""
                )

            # add lwc
            lwc_path = os.path.join(path, "lwc")
            os.mkdir(lwc_path)

            # add lwc linting files (not included in zip)
            lwc_ignored_files = [".eslintrc.json", "jsconfig.json"]
            for lwc_ignored_file in lwc_ignored_files:
                touch(os.path.join(lwc_path, lwc_ignored_file))

            # add lwc component
            lwc_component_path = os.path.join(lwc_path, "myComponent")
            os.mkdir(lwc_component_path)

            # add lwc component files included in zip (in alphabetical order)
            lwc_component_files = [
                {"name": "myComponent.html"},
                {"name": "myComponent.js"},
                {
                    "name": "myComponent.js-meta.xml",
                    "body:": """<?xml version="1.0" encoding="UTF-8"?>
<LightningComponentBundle xmlns="http://soap.sforce.com/2006/04/metadata" fqn="myComponent">
    <apiVersion>45.0</apiVersion>
    <isExposed>false</isExposed>
</LightningComponentBundle>""",
                },
                {"name": "myComponent.svg"},
                {"name": "myComponent.css"},
            ]
            for lwc_component_file in lwc_component_files:
                with open(
                    os.path.join(lwc_component_path, lwc_component_file.get("name")),
                    "w",
                ) as f:
                    if lwc_component_file.get("body") is not None:
                        f.write(lwc_component_file.get("body"))

            # add lwc component files not included in zip
            for lwc_ignored_file in lwc_ignored_files:
                touch(os.path.join(lwc_component_path, lwc_ignored_file))

            # add lwc component sub-directory and files not included in zip
            lwc_component_test_path = os.path.join(lwc_component_path, "__tests__")
            os.mkdir(lwc_component_test_path)
            touch(os.path.join(lwc_component_test_path, "test.js"))

            # add classes
            classes_path = os.path.join(path, "classes")
            os.mkdir(classes_path)
            class_files = [
                {
                    "name": "MyClass.cls-meta.xml",
                    "body": """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>45.0</apiVersion>
    <status>Active</status>
</ApexClass>
""",
                },
                {"name": "MyClass.cls"},
            ]
            for class_file in class_files:
                with open(os.path.join(classes_path, class_file.get("name")), "w") as f:
                    if class_file.get("body") is not None:
                        f.write(class_file.get("body"))

            # add objects
            objects_path = os.path.join(path, "objects")
            os.mkdir(objects_path)
            object_file_names = ["Account.object", "Contact.object", "CustomObject__c"]
            object_file_names.sort()
            for object_file_name in object_file_names:
                touch(os.path.join(objects_path, object_file_name))

            # add sub-directory of objects (that doesn't really exist)
            objects_sub_path = os.path.join(objects_path, "does-not-exist-in-schema")
            os.mkdir(objects_sub_path)
            touch(os.path.join(objects_sub_path, "some.file"))

            # test
            builder = MetadataPackageZipBuilder(
                path=path,
                options={
                    "namespace_tokenize": "ns",
                    "namespace_inject": "ns",
                    "namespace_strip": "ns",
                },
                context=task_context,
            )

            # make sure result can be read as a zipfile
            result = builder.as_base64()
            zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(result)), "r")
            assert set(zf.namelist()) == {
                "package.xml",
                "lwc/myComponent/myComponent.html",
                "lwc/myComponent/myComponent.js",
                "lwc/myComponent/myComponent.js-meta.xml",
                "lwc/myComponent/myComponent.svg",
                "lwc/myComponent/myComponent.css",
                "classes/MyClass.cls-meta.xml",
                "classes/MyClass.cls",
                "objects/Account.object",
                "objects/Contact.object",
                "objects/CustomObject__c",
                "objects/does-not-exist-in-schema/some.file",
            }
            zf.close()

    def test_add_files_to_package(self, task_context):
        with temporary_dir() as path:
            expected = []

            # add package.xml
            with open(os.path.join(path, "package.xml"), "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <version>45.0</version>
</Package>"""
                )
                expected.append("package.xml")

            # add lwc
            lwc_path = os.path.join(path, "lwc")
            os.mkdir(lwc_path)

            # add lwc linting files (not included in zip)
            lwc_ignored_files = [".eslintrc.json", "jsconfig.json"]
            for lwc_ignored_file in lwc_ignored_files:
                touch(os.path.join(lwc_path, lwc_ignored_file))

            # add lwc component
            lwc_component_path = os.path.join(lwc_path, "myComponent")
            os.mkdir(lwc_component_path)

            # add lwc component files included in zip (in alphabetical order)
            lwc_component_files = [
                {"name": "myComponent.html"},
                {"name": "myComponent.js"},
                {
                    "name": "myComponent.js-meta.xml",
                    "body:": """<?xml version="1.0" encoding="UTF-8"?>
<LightningComponentBundle xmlns="http://soap.sforce.com/2006/04/metadata" fqn="myComponent">
    <apiVersion>45.0</apiVersion>
    <isExposed>false</isExposed>
</LightningComponentBundle>""",
                },
                {"name": "myComponent.svg"},
                {"name": "myComponent.css"},
            ]
            for lwc_component_file in lwc_component_files:
                with open(
                    os.path.join(lwc_component_path, lwc_component_file.get("name")),
                    "w",
                ) as f:
                    if lwc_component_file.get("body") is not None:
                        f.write(lwc_component_file.get("body"))
                    expected.append(f"lwc/myComponent/{lwc_component_file.get('name')}")

            # add lwc component files not included in zip
            for lwc_ignored_file in lwc_ignored_files:
                touch(os.path.join(lwc_component_path, lwc_ignored_file))

            # add lwc component sub-directory and files not included in zip
            lwc_component_test_path = os.path.join(lwc_component_path, "__tests__")
            os.mkdir(lwc_component_test_path)
            touch(os.path.join(lwc_component_test_path, "test.js"))

            # add classes
            classes_path = os.path.join(path, "classes")
            os.mkdir(classes_path)
            class_files = [
                {
                    "name": "MyClass.cls-meta.xml",
                    "body": """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>45.0</apiVersion>
    <status>Active</status>
</ApexClass>
""",
                },
                {"name": "MyClass.cls"},
            ]
            for class_file in class_files:
                with open(os.path.join(classes_path, class_file.get("name")), "w") as f:
                    if class_file.get("body") is not None:
                        f.write(class_file.get("body"))
                    expected.append(f"classes/{class_file.get('name')}")

            # add objects
            objects_path = os.path.join(path, "objects")
            os.mkdir(objects_path)
            object_file_names = ["Account.object", "Contact.object", "CustomObject__c"]
            object_file_names.sort()
            for object_file_name in object_file_names:
                with open(os.path.join(objects_path, object_file_name), "w"):
                    expected.append(f"objects/{object_file_name}")

            # add sub-directory of objects (that doesn't really exist)
            objects_sub_path = os.path.join(objects_path, "does-not-exist-in-schema")
            os.mkdir(objects_sub_path)
            with open(os.path.join(objects_sub_path, "some.file"), "w"):
                expected.append("objects/does-not-exist-in-schema/some.file")

            # test
            builder = MetadataPackageZipBuilder(context=task_context)

            expected_set = set(expected)
            builder._add_files_to_package(path)
            actual_set = set(builder.zf.namelist())
            assert expected_set == actual_set

    def test_include_directory(self, task_context):
        builder = MetadataPackageZipBuilder(context=task_context)

        # include root directory
        assert builder._include_directory([]) is True

        # not include lwc directory
        assert builder._include_directory(["lwc"]) is True

        # include any lwc sub-directory (i.e. lwc component directory)
        assert builder._include_directory(["lwc", "myComponent"]) is True
        assert builder._include_directory(["lwc", "lwc"]) is True
        assert (
            builder._include_directory(["lwc", "myComponent", "sub-1", "sub-2"]) is True
        )

        # don't include __tests__ within lwc components
        assert not builder._include_directory(["lwc", "myComponent", "__tests__"])
        assert not builder._include_directory(["lwc", "myComponent", "__mocks__"])

        # include any non-lwc directory
        assert builder._include_directory(["not-lwc"]) is True
        assert builder._include_directory(["classes"]) is True
        assert builder._include_directory(["objects"]) is True

        # include any sub_* directory of a non-lwc directory
        assert builder._include_directory(["not-lwc", "sub-1"]) is True
        assert builder._include_directory(["not-lwc", "sub-1", "sub-2"]) is True

    def test_include_file(self, task_context):
        builder = MetadataPackageZipBuilder(context=task_context)

        lwc_component_directories = [
            ["lwc"],
            ["lwc", "myComponent"],
            ["lwc", "myComponent", "sub-1"],
            ["lwc", "myComponent", "sub-2"],
        ]
        non_lwc_component_directories = [
            [],
            ["classes"],
            ["objects", "sub-1"],
            ["objects", "sub-1", "sub-2"],
        ]

        # file endings in lwc component whitelist
        for file_ending in [".js", ".js-meta.xml", ".html", ".css", ".svg"]:
            for d in lwc_component_directories:
                assert builder._include_file(d, "file_name" + file_ending)
            for d in non_lwc_component_directories:
                assert builder._include_file(d, "file_name" + file_ending)

        # file endings not in lwc component whitelist
        for file_ending in ["", ".json", ".xml", ".cls", ".cls-meta.xml", ".object"]:
            for d in lwc_component_directories:
                assert not builder._include_file(d, "file_name" + file_ending)
            for d in non_lwc_component_directories:
                assert builder._include_file(d, "file_name" + file_ending)

    def test_removes_feature_parameters_from_unlocked_package(self, task_context):
        with temporary_dir() as path:
            pathlib.Path(path, "package.xml").write_text(
                """<?xml version="1.0" encoding="utf-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <name>FeatureParameterInteger</name>
    </types>
</Package>"""
            )
            featureParameters = pathlib.Path(path, "featureParameters")
            featureParameters.mkdir()
            (featureParameters / "test.featureParameterInteger").touch()
            builder = MetadataPackageZipBuilder(
                path=path, options={"package_type": "Unlocked"}, context=task_context
            )
            assert (
                "featureParameters/test.featureParameterInteger"
                not in builder.zf.namelist()
            )
            package_xml = builder.zf.read("package.xml")
            assert b"FeatureParameterInteger" not in package_xml


class TestCreatePackageZipBuilder:
    def test_init__missing_name(self):
        with pytest.raises(ValueError):
            CreatePackageZipBuilder(None, "43.0")

    def test_init__missing_api_version(self):
        with pytest.raises(ValueError):
            CreatePackageZipBuilder("TestPackage", None)


class TestInstallPackageZipBuilder:
    def test_init__missing_namespace(self):
        with pytest.raises(ValueError):
            InstallPackageZipBuilder(None, "1.0")

    def test_init__missing_version(self):
        with pytest.raises(ValueError):
            InstallPackageZipBuilder("testns", None)


class TestDestructiveChangesZipBuilder:
    def test_call(self):
        builder = DestructiveChangesZipBuilder("", "1.0")
        names = builder.zf.namelist()
        assert "package.xml" in names
        assert "destructiveChanges.xml" in names


class TestUninstallPackageZipBuilder:
    def test_init__missing_namespace(self):
        with pytest.raises(ValueError):
            UninstallPackageZipBuilder(None, "1.0")

    def test_call(self):
        builder = UninstallPackageZipBuilder("testns", "1.0")
        assert "destructiveChanges.xml" in builder.zf.namelist()



import typing as T

from simple_salesforce import Salesforce

from cumulusci.utils.http.multi_request import CompositeParallelSalesforce
from cumulusci.utils.iterators import partition


class ObjectCount(T.NamedTuple):
    counts: T.Dict[str, int]
    transport_errors: T.Sequence[dict]  # e.g. couldn't reach server
    salesforce_errors: T.Sequence[dict]  # e.g. 404, 401


def count_sobjects(sf: Salesforce, objs: T.Sequence[str]) -> ObjectCount:
    """Quickly count SObjects using SOQL and Parallelization"""
    with CompositeParallelSalesforce(sf, max_workers=8, chunk_size=5) as cpsf:
        responses, transport_errors = cpsf.do_composite_requests(
            (
                {
                    "method": "GET",
                    "url": f"/services/data/v{sf.sf_version}/query/?q=select count() from {obj}",
                    "referenceId": f"ref{obj}",
                }
                for obj in objs
            )
        )
        salesforce_errors, successes = partition(
            lambda response: response["httpStatusCode"] == 200, responses
        )

        transport_errors = tuple(error._asdict() for error in transport_errors)

    successes = list(successes)

    def removeprefix(mainstr: str, prefix: str):
        # Until Python 3.9 is minimum supported
        a, b = mainstr[0 : len(prefix)], mainstr[len(prefix) :]
        assert a == prefix
        return b

    ret = {
        removeprefix(response["referenceId"], "ref"): response["body"]["totalSize"]
        for response in successes
    }
    return ObjectCount(ret, transport_errors, tuple(salesforce_errors))



import code
import contextlib
import pdb
import runpy
import sys
import traceback

import click
import requests
import rich
from rich.console import Console
from rich.markup import escape

import cumulusci
from cumulusci.core.debug import set_debug_mode
from cumulusci.core.exceptions import CumulusCIUsageError
from cumulusci.utils import get_cci_upgrade_command
from cumulusci.utils.http.requests_utils import init_requests_trust
from cumulusci.utils.logging import tee_stdout_stderr

from .error import error
from .flow import flow
from .logger import get_tempfile_logger, init_logger
from .org import org
from .plan import plan
from .project import project
from .robot import robot
from .runtime import CliRuntime, pass_runtime
from .service import service
from .task import task
from .utils import (
    check_latest_version,
    get_installed_version,
    get_latest_final_version,
    warn_if_no_long_paths,
)

SUGGEST_ERROR_COMMAND = (
    """Run this command for more information about debugging errors: cci error --help"""
)

USAGE_ERRORS = (CumulusCIUsageError, click.UsageError)


#
# Root command
#
def main(args=None):
    """Main CumulusCI CLI entry point.

    This runs as the first step in processing any CLI command.

    This wraps the `click` library in order to do some initialization and centralized error handling.
    """
    with contextlib.ExitStack() as stack:
        args = args or sys.argv

        # (If enabled) set up requests to validate certs using system CA certs instead of certifi
        init_requests_trust()

        # Check for updates _unless_ we've been asked to output JSON,
        # or if we're going to check anyway as part of the `version` command.
        is_version_command = len(args) > 1 and args[1] == "version"
        if "--json" not in args and not is_version_command:
            check_latest_version()

        # Only create logfiles for commands that are not `cci error`
        is_error_command = len(args) > 2 and args[1] == "error"
        tempfile_path = None
        if not is_error_command:
            logger, tempfile_path = get_tempfile_logger()
            stack.enter_context(tee_stdout_stderr(args, logger, tempfile_path))

        debug = "--debug" in args
        if debug:
            args.remove("--debug")

        with set_debug_mode(debug):
            try:
                runtime = CliRuntime(load_keychain=False)
            except Exception as e:
                handle_exception(e, is_error_command, tempfile_path, debug)
                sys.exit(1)

            runtime.check_cumulusci_version()
            should_show_stacktraces = runtime.universal_config.cli__show_stacktraces

            init_logger(debug=debug)
            # Hand CLI processing over to click, but handle exceptions
            try:
                cli(args[1:], standalone_mode=False, obj=runtime)
            except click.Abort:  # Keyboard interrupt
                console = Console()
                show_debug_info() if debug else console.print("\n[red bold]Aborted!")
                sys.exit(1)
            except Exception as e:
                if debug:
                    show_debug_info()
                else:
                    handle_exception(
                        e,
                        is_error_command,
                        tempfile_path,
                        should_show_stacktraces,
                    )
                sys.exit(1)


def handle_exception(
    error,
    is_error_cmd,
    logfile_path,
    should_show_stacktraces=False,
):
    """Displays error of appropriate message back to user, prompts user to investigate further
    with `cci error` commands, and writes the traceback to the latest logfile.
    """
    error_console = Console(stderr=True)
    if isinstance(error, requests.exceptions.ConnectionError):
        connection_error_message(error_console)
    elif isinstance(error, click.ClickException):
        error_console.print(f"[red bold]Error: {escape(error.format_message())}")
    else:
        # We call str ourselves to make Typeguard shut up.
        error_console.print(f"[red bold]Error: {escape(str(error))}")
    # Only suggest gist command if it wasn't run
    if not is_error_cmd:
        error_console.print(f"[yellow]{SUGGEST_ERROR_COMMAND}")

    # This is None if we're handling an exception for a `cci error` command.
    if logfile_path:
        with open(logfile_path, "a") as log_file:
            traceback.print_exc(file=log_file)  # log stacktrace silently

    if should_show_stacktraces and not isinstance(error, USAGE_ERRORS):
        error_console.print_exception()


def connection_error_message(console: Console):
    message = (
        "We encountered an error with your internet connection. "
        "Please check your connection and try the last cci command again."
    )
    console.print(f"[red bold]{message}")


def show_debug_info():
    """Displays the traceback and opens pdb"""
    traceback.print_exc()
    pdb.post_mortem()


def show_version_info():
    console = rich.get_console()
    console.print(f"CumulusCI version: {cumulusci.__version__} ({sys.argv[0]})")
    console.print(f"Python version: {sys.version.split()[0]} ({sys.executable})")
    console.print()
    warn_if_no_long_paths(console=console)

    current_version = get_installed_version()
    latest_version = get_latest_final_version()

    if not latest_version > current_version:
        console.print("You have the latest version of CumulusCI :sun_behind_cloud:\n")
        display_release_notes_link(str(latest_version))
        return

    console.print(
        f"[yellow]There is a newer version of CumulusCI available: {str(latest_version)}"
    )
    console.print(f"To upgrade, run `{get_cci_upgrade_command()}`")
    display_release_notes_link(str(latest_version))


def display_release_notes_link(latest_version: str) -> None:
    """Provide a link to the latest CumulusCI Release Notes"""
    release_notes_link = (
        f"https://github.com/SFDO-Tooling/CumulusCI/releases/tag/v{latest_version}"
    )
    console = rich.get_console()
    console.print(
        f"See the latest CumulusCI Release Notes: [link={release_notes_link}]{release_notes_link}[/link]"
    )


def version_info_wrapper(
    ctx: click.Context, param: click.Parameter, value: bool
) -> None:
    if not value:
        return
    show_version_info()
    ctx.exit()


@click.group("main", help="")
@click.option(  # based on https://click.palletsprojects.com/en/8.1.x/options/#callbacks-and-eager-options
    "--version",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    help="Show the version and exit.",
    callback=version_info_wrapper,
)
def cli():
    """Top-level `click` command group."""


@cli.command(name="version", help="Print the current version of CumulusCI")
def version():
    show_version_info()


@cli.command(name="shell", help="Drop into a Python shell")
@click.option("--script", help="Path to a script to run", type=click.Path())
@click.option("--python", help="Python code to run directly")
@pass_runtime(require_project=False, require_keychain=True)
def shell(runtime, script=None, python=None):
    # alias for backwards-compatibility
    variables = {
        "config": runtime,
        "runtime": runtime,
        "project_config": runtime.project_config,
    }

    if script:
        if python:
            raise click.UsageError("Cannot specify both --script and --python")
        runpy.run_path(script, init_globals=variables)
    elif python:
        exec(python, variables)
    else:
        code.interact(local=variables)


# Top Level Groups

cli.add_command(error)
cli.add_command(project)
cli.add_command(org)
cli.add_command(service)
cli.add_command(task)
cli.add_command(flow)
cli.add_command(plan)
cli.add_command(robot)



import pytest
import responses
from requests.models import Response
from requests.sessions import Session
from responses import matchers

from cumulusci.core.config import ServiceConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.metadeploy.api import MetaDeployAPI, make_api_session

pytestmark = pytest.mark.metadeploy
DEFAULT_REST_URL: str = "https://metadeploy.example.com/api/rest"
DEFAULT_TOKEN: str = "37b003aee9bdd744a6618e9fe12"


@pytest.fixture
def default_api_client():
    return MetaDeployAPI(
        metadeploy_service=ServiceConfig(
            {"url": DEFAULT_REST_URL, "token": DEFAULT_TOKEN}
        )
    )


@pytest.fixture
def auth_header_matcher():
    return matchers.header_matcher({"Authorization": f"token {DEFAULT_TOKEN}"})


@responses.activate
def test_make_api(auth_header_matcher):
    expected_result: dict = {"data": [1]}
    responses.add(
        "GET",
        f"{DEFAULT_REST_URL}/resource",
        json=expected_result,
        match=[auth_header_matcher],
    )
    responses.add(
        "POST",
        f"{DEFAULT_REST_URL}/resource",
        json=expected_result,
        match=[auth_header_matcher],
    )
    service: ServiceConfig = ServiceConfig(
        {"url": f"{DEFAULT_REST_URL}", "token": DEFAULT_TOKEN}
    )
    api_session: Session = make_api_session(service)
    resp: Response = api_session.get("/resource")
    result: dict = resp.json()
    assert result == expected_result
    _ = api_session.get(f"{DEFAULT_REST_URL}/resource")

    assert responses.assert_call_count(f"{DEFAULT_REST_URL}/resource", 2) is True


def test_find_product(default_api_client, auth_header_matcher):
    params = {"repo_url": "https://api.github.com/repos/TestOwner/TestRepo"}
    payload = {
        "data": [
            {
                "id": "abcdef",
                "url": "https://EXISTING_PRODUCT",
                "slug": "existing",
            }
        ]
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=f"{DEFAULT_REST_URL}/products",
            match=(auth_header_matcher, matchers.query_param_matcher(params)),
            json=payload,
        )

        product: dict = default_api_client.find_product(params["repo_url"])
    assert product == payload["data"][0]


def test_find_product__api_err(default_api_client, auth_header_matcher):
    params = {"repo_url": "https://api.github.com/repos/TestOwner/MISSING"}
    payload = {"result": []}

    with pytest.raises(CumulusCIException, match="unexpected response"):
        with responses.RequestsMock() as rsps:
            rsps.add(
                method=responses.GET,
                url=f"{DEFAULT_REST_URL}/products",
                match=(auth_header_matcher, matchers.query_param_matcher(params)),
                json=payload,
            )

            default_api_client.find_product(params["repo_url"])


def test_create_plan(default_api_client, auth_header_matcher):
    payload = {"url": f"{DEFAULT_REST_URL}/plans/1"}
    plan = {
        "is_listed": True,
        "plan_template": "https://foo.example.com/1",
        "post_install_message_additional": "",
        "preflight_message_additional": "",
        "steps": [],
        "supported_orgs": "persistent",
        "tier": "primary",
        "title": "Install Tubulator 1.0",
        "version": f"{DEFAULT_REST_URL}/version/1",
        "visible_to": None,
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=f"{DEFAULT_REST_URL}/plans",
            match=(auth_header_matcher, matchers.json_params_matcher(plan)),
            status=201,
            json=payload,
        )
        result = default_api_client.create_plan(plan)
    assert result == payload


def test_find_version(default_api_client, auth_header_matcher):
    query = {"product": "bDcaVz", "label": "1.0"}
    payload = {"data": [{"url": "http://EXISTING_VERSION"}]}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=f"{DEFAULT_REST_URL}/versions",
            match=(auth_header_matcher, matchers.query_param_matcher(query)),
            json=payload,
        )

        result = default_api_client.find_version(query)

    assert result == payload["data"][0]


def test_find_version_missing_none(default_api_client, auth_header_matcher):
    query = {"product": "bDcaVz", "label": "1.0"}
    payload = {"data": []}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=f"{DEFAULT_REST_URL}/versions",
            match=(auth_header_matcher, matchers.query_param_matcher(query)),
            json=payload,
        )

        result = default_api_client.find_version(query)

    assert result is None


def test_create_version(default_api_client, auth_header_matcher):
    version = {
        "product": f"{DEFAULT_REST_URL}/product/foo",
        "label": "1.0",
        "description": "",
        "is_production": True,
        "commit_ish": "tag_or_sha",
        "is_listed": False,
    }
    payload = {"url": f"http://{DEFAULT_REST_URL}/versions/1", "id": 1}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=f"{DEFAULT_REST_URL}/versions",
            match=(auth_header_matcher, matchers.json_params_matcher(version)),
            status=201,
            json=payload,
        )

        version = default_api_client.create_version(version)

    assert version == payload


def test_publish_version(default_api_client, auth_header_matcher):
    vers = {"is_listed": True}
    pk = "XG4pVvD"
    payload = {"url": f"http://{DEFAULT_REST_URL}/versions/{pk}", "id": pk}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PATCH,
            url=f"{DEFAULT_REST_URL}/versions/{pk}",
            match=(auth_header_matcher, matchers.json_params_matcher(vers)),
            json=payload,
        )

        updated_version = default_api_client.update_version(pk)
    assert payload == updated_version


def test_find_plan_template(default_api_client, auth_header_matcher):
    query = {"product": "bDcaVz", "name": "Shinra"}
    payload = {"data": [{"url": "http://existing_template"}]}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=f"{DEFAULT_REST_URL}/plantemplates",
            match=(auth_header_matcher, matchers.query_param_matcher(query)),
            json=payload,
        )

        template = default_api_client.find_plan_template(query)

    assert template == payload["data"][0]


def test_create_plan_template(default_api_client, auth_header_matcher):
    template = {
        "name": "plan_name",
        "product": "product",
        "preflight_message": "",
        "post_install_message": "",
        "error_message": "oops",
    }
    payload = {"url": "http://EXISTING_VERSION"}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=f"{DEFAULT_REST_URL}/plantemplates",
            match=(auth_header_matcher, matchers.json_params_matcher(template)),
            status=201,
            json=payload,
        )

        created_template = default_api_client.create_plan_template(template)

    assert created_template == payload


def test_create_plan_slug(default_api_client, auth_header_matcher):
    slug = {"slug": "plan_name", "parent": "product"}
    payload = {"url": "http://EXISTING_VERSION"}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=f"{DEFAULT_REST_URL}/planslug",
            match=(auth_header_matcher, matchers.json_params_matcher(slug)),
            status=201,
            json=payload,
        )

        created_slug = default_api_client.create_plan_slug(slug)

    assert created_slug == payload


def test_publish_labels(default_api_client, auth_header_matcher):
    lang = "en_us"
    labels = {
        "title": "name of product",
        "error_message": "shown after failed installation (markdown)",
    }
    payload = {}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PATCH,
            url=f"{DEFAULT_REST_URL}/translations/{lang}",
            match=(auth_header_matcher, matchers.json_params_matcher(labels)),
            json=payload,
        )

        updated_txn = default_api_client.update_lang_translation(lang, labels)
    assert payload == updated_txn



import csv
from collections import defaultdict
from pathlib import PurePosixPath
from typing import List, Optional, Union
from unittest.mock import Mock
from zipfile import ZipFile

from github3.repos.repo import Repository
from pydantic import BaseModel

from cumulusci.core.dependencies.dependencies import (
    Dependency,
    GitHubDynamicDependency,
    parse_dependencies,
)
from cumulusci.core.dependencies.github import (
    get_package_data,
    get_remote_project_config,
    get_repo,
)
from cumulusci.core.dependencies.resolvers import get_static_dependencies
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.github.base import BaseGithubTask
from cumulusci.utils import download_extract_github_from_repo
from cumulusci.utils.version_strings import LooseVersion
from cumulusci.utils.xml import metadata_tree


class Package(BaseModel):
    repo: Optional[Union[Repository, Mock]]
    package_name: str
    namespace: str
    prefix_release: str

    def __hash__(self) -> int:
        return self.repo.__hash__()

    class Config:
        arbitrary_types_allowed = True


class PackageVersion(BaseModel):
    package: Package
    version: LooseVersion

    class Config:
        arbitrary_types_allowed = True


class SObjectDetail(BaseModel):
    version: PackageVersion
    api_name: str
    label: str
    description: str


class FieldDetail(BaseModel):
    version: PackageVersion
    sobject: str
    api_name: str
    label: str
    type: str
    help_text: str
    description: str
    valid_values: str


# "Version number" used to represent a prerelease.
PRERELEASE_SIGIL = LooseVersion("100000001.0")


class GenerateDataDictionary(BaseGithubTask):
    task_docs = """
    Generate a data dictionary for the project by walking all GitHub releases.
    The data dictionary is output as two CSV files.
    One, in `object_path`, includes

    - Object Label
    - Object API Name
    - Object Description
    - Version Introduced

    with one row per packaged object.

    The other, in `field_path`, includes

    - Object Label
    - Object API Name
    - Field Label
    - Field API Name
    - Field Type
    - Valid Picklist Values
    - Help Text
    - Field Description
    - Version Introduced
    - Version Picklist Values Last Changed
    - Version Help Text Last Changed

    Both MDAPI and SFDX format releases are supported.
    """

    task_options = {
        "object_path": {
            "description": "Path to a CSV file to contain an sObject-level data dictionary."
        },
        "field_path": {
            "description": "Path to a CSV file to contain an field-level data dictionary."
        },
        "include_dependencies": {
            "description": "Process all of the GitHub dependencies of this project and "
            "include their schema in the data dictionary.",
            "default": True,
        },
        "additional_dependencies": {
            "description": "Include schema from additional GitHub repositories that "
            "are not explicit dependencies of this project to build a unified data dictionary. "
            "Specify as a list of dicts as in project__dependencies in cumulusci.yml. Note: only "
            "repository dependencies are supported."
        },
        "include_prerelease": {
            "description": "Treat the current branch as containing prerelease schema, "
            "and included it as Prerelease in the data dictionary. NOTE: this option "
            "cannot be used with `additional_dependencies` or `include_dependencies`."
        },
        "include_protected_schema": {
            "description": "Include Custom Objects, Custom Settings, and Custom Metadata "
            "Types that are marked as Protected. Defaults to False."
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        if self.options.get("object_path") is None:
            self.options[
                "object_path"
            ] = f"{self.project_config.project__name} Objects.csv"

        if self.options.get("field_path") is None:
            self.options[
                "field_path"
            ] = f"{self.project_config.project__name} Fields.csv"

        include_dependencies = self.options.get("include_dependencies")
        self.options["include_dependencies"] = process_bool_arg(
            True if include_dependencies is None else include_dependencies
        )

        if "additional_dependencies" in self.options:
            additional_deps = parse_dependencies(
                self.options["additional_dependencies"]
            )
            if not all(isinstance(d, GitHubDynamicDependency) for d in additional_deps):
                raise TaskOptionsError(
                    "Only GitHub dependencies are currently supported."
                )

        self.options["include_prerelease"] = process_bool_arg(
            self.options.get("include_prerelease") or False
        )

        if self.options["include_prerelease"]:
            if self.options.get("additional_dependencies"):
                raise TaskOptionsError(
                    "The additional_dependencies option cannot be used with include_prerelease."
                )

            if self.options["include_dependencies"]:
                self.logger.info(
                    "Setting include_prerelease prohibits include_dependencies; setting include_dependencies to False"
                )
                self.options["include_dependencies"] = False

        self.options["include_protected_schema"] = process_bool_arg(
            self.options.get("include_protected_schema") or False
        )

    def _get_repo_dependencies(
        self, dependencies: List[GitHubDynamicDependency]
    ) -> List[Package]:
        """Return a list of Package objects representing all of the GitHub repositories
        in this project's dependency tree. Ignore all non-GitHub dependencies."""
        github_deps = set()
        packages = []

        def log_github(some_dep: Dependency):
            if isinstance(some_dep, GitHubDynamicDependency):
                github_deps.add(some_dep)

            return True

        _ = get_static_dependencies(
            self.project_config,
            dependencies,
            resolution_strategy="production",
            filter_function=log_github,
        )

        for dependency in github_deps:
            repo = get_repo(dependency.github, self.project_config)
            config = get_remote_project_config(repo, dependency.ref)
            package_name, namespace = get_package_data(config)
            packages.append(
                Package(
                    repo=repo,
                    package_name=package_name,
                    namespace=f"{namespace}__" if namespace else "",
                    prefix_release=config.project__git__prefix_release or "release/",
                )
            )

        return packages

    def _run_task(self):
        self.logger.info("Starting data dictionary generation")

        self._init_schema()

        namespace = self.project_config.project__package__namespace
        if namespace:
            namespace = f"{namespace}__"
        repos = [
            Package(
                repo=self.get_repo(),
                package_name=self.project_config.project__package__name,
                namespace=namespace,
                prefix_release=self.project_config.project__git__prefix_release,
            )
        ]

        # Find all of our dependencies, if we're processing dependencies.
        dependencies = []
        if (
            self.options["include_dependencies"]
            and self.project_config.project__dependencies
        ):
            parsed_deps = parse_dependencies(self.project_config.project__dependencies)
            dependencies.extend(
                d for d in parsed_deps if isinstance(d, GitHubDynamicDependency)
            )
        if "additional_dependencies" in self.options:
            # init_options() required these to all be GitHubDynamicDependencies
            dependencies.extend(
                parse_dependencies(self.options["additional_dependencies"])
            )

        if dependencies:
            repos.extend(self._get_repo_dependencies(dependencies))

        for package in repos:
            self._walk_releases(package)

        self._write_results()

    def _init_schema(self):
        """Initialize the structure used for schema storage."""
        self.sobjects = defaultdict(list)
        self.fields = defaultdict(list)
        self.package_versions = defaultdict(list)
        self.omit_sobjects = set()

    def _walk_releases(self, package: Package):
        """Traverse all of the releases in this project's repository and process
        each one matching our tag (not draft/prerelease) to generate the data dictionary."""
        for release in package.repo.releases():
            # Skip this release if any are true:
            # It is a draft release
            # It is prerelease (managed beta)
            # This release's tag does not have the expected prefix,
            # meaning we don't know its version number
            if (
                release.draft
                or release.prerelease
                or not release.tag_name.startswith(package.prefix_release)
            ):
                continue

            zip_file = download_extract_github_from_repo(
                package.repo, ref=release.tag_name
            )
            version = PackageVersion(
                package=package,
                version=self._version_from_tag_name(
                    release.tag_name, package.prefix_release
                ),
            )
            self.package_versions[package].append(version.version)
            self.logger.info(
                f"Analyzing {package.package_name} version {version.version}"
            )

            self._process_zipfile(zip_file, version)

        # If we are asked to process a prerelease, do so.
        if self.options["include_prerelease"]:
            # package.repo is guaranteed to be our repo (via _init_options())
            zip_file = download_extract_github_from_repo(
                package.repo, ref=self.project_config.repo_branch
            )
            version = PackageVersion(
                package=package,
                version=PRERELEASE_SIGIL,
            )
            self.package_versions[package].append(version.version)
            self.logger.info(
                f"Analyzing {package.package_name} prerelease from {self.project_config.repo_branch}"
            )

            self._process_zipfile(zip_file, version)

    def _process_zipfile(self, zip_file: ZipFile, version: PackageVersion):
        if "src/objects/" in zip_file.namelist():
            # MDAPI format
            self._process_mdapi_release(zip_file, version)
        elif "sfdx-project.json" in zip_file.namelist():
            # SFDX format
            # Note: we check MDAPI first, because many
            # CumulusCI projects contain an sfdx-project.json
            # along with MDAPI source.
            self._process_sfdx_release(zip_file, version)

    def _process_mdapi_release(self, zip_file: ZipFile, version: PackageVersion):
        """Process an MDAPI ZIP file for objects and fields"""
        for f in zip_file.namelist():
            path = PurePosixPath(f)
            if path.parent == PurePosixPath("src/objects") and path.suffix == ".object":
                sobject_name = path.stem
                if sobject_name.count("__") == 1:
                    sobject_name = f"{version.package.namespace}{sobject_name}"

                self._process_object_element(
                    sobject_name, metadata_tree.fromstring(zip_file.read(f)), version
                )

    def _should_process_object(
        self,
        namespace: str,
        sobject_name: str,
        element: Optional[metadata_tree.MetadataElement],
    ):
        """Determine if we should track this object in the object dictionary.
        Fields may be included regardless.

        We don't process custom objects owned by other namespaces.
        It's fine for Package A to package fields on an object owned by Package B.
        We'll record the fields on their owning package (B) and the object on its (A)."""

        return (
            (
                sobject_name.endswith("__c")
                or sobject_name.endswith("__mdt")
                or sobject_name.endswith("__e")
                or sobject_name.count("__") == 0
            )
            and sobject_name.startswith(namespace)
            and (
                not self._is_protected_object(element)
                or self.options["include_protected_schema"]
            )
        )

    def _should_process_object_fields(
        self, sobject_name: str, element: Optional[metadata_tree.MetadataElement]
    ) -> bool:
        """Determine if we should track fields from this object in the field dictionary.
        The object itself may or may not be included.

        We will track any field on a custom object, Platform Event,
        Custom Setting, Custom Metadata Type, or standard entity.
        We will not track any object's fields that are protected
        in this package version."""
        return (
            sobject_name.endswith("__c")
            or sobject_name.endswith("__mdt")
            or sobject_name.endswith("__e")
            or sobject_name.count("__") == 0
        ) and (
            not self._is_protected_object(element)
            or self.options["include_protected_schema"]
        )

    def _is_protected_object(
        self, element: Optional[metadata_tree.MetadataElement]
    ) -> bool:
        if not element:
            return False

        # customSettingsVisibility is used in older API versions.
        visibility_elem = element.find("customSettingsVisibility")
        if visibility_elem:
            return visibility_elem.text != "Public"

        visibility_elem = element.find("visibility")
        if visibility_elem:
            return visibility_elem.text != "Public"

        return False

    def _process_sfdx_release(self, zip_file: ZipFile, version: PackageVersion):
        """Process an SFDX ZIP file for objects and fields"""
        for f in zip_file.namelist():
            path = PurePosixPath(f)
            # Be flexible about processing directories in SFDX context.
            # This may not be optimal if the repo contains multiple
            # 2GP package subdirectories.
            if "objects/" in f and not f.startswith("unpackaged/"):
                if path.suffixes == [".object-meta", ".xml"]:
                    sobject_name = path.name[: -len(".object-meta.xml")]
                    if sobject_name.count("__") == 1:
                        sobject_name = f"{version.package.namespace}{sobject_name}"

                    element = metadata_tree.fromstring(zip_file.read(f))

                    if self._should_process_object(
                        version.package.namespace, sobject_name, element
                    ):
                        self._process_object_element(sobject_name, element, version)
                    else:
                        # If this is an object type from which we shouldn't process any fields,
                        # track it in omit_sobjects so we can drop any fields later if we don't have
                        # the right information at time of processing.

                        # Note that the owning object may be in a dependency package, so we won't find it below.
                        if not self._should_process_object_fields(
                            sobject_name, element
                        ):
                            self.omit_sobjects.add(sobject_name)
                elif path.suffixes == [".field-meta", ".xml"]:
                    # To get the sObject name, we need to remove the `/fields/SomeField.field-meta.xml`
                    # and take the last path component

                    # Find the sObject metadata file
                    sobject_name = f"{path.parent.parent.stem}"
                    sobject_file = str(
                        path.parent.parent / f"{sobject_name}.object-meta.xml"
                    )
                    if sobject_name.count("__") == 1:
                        sobject_name = f"{version.package.namespace}{sobject_name}"

                    # If the object-meta file is locatable, load it so we can check
                    # if this is a Custom Setting.
                    if sobject_file in zip_file.namelist():
                        object_entity = metadata_tree.fromstring(
                            zip_file.read(sobject_file)
                        )
                    else:
                        object_entity = None

                    if self._should_process_object_fields(sobject_name, object_entity):
                        self._process_field_element(
                            sobject_name,
                            metadata_tree.fromstring(zip_file.read(f)),
                            version,
                        )

    def _process_object_element(
        self,
        sobject_name: str,
        element: metadata_tree.MetadataElement,
        version: PackageVersion,
    ):
        """Process a <CustomObject> metadata entity, whether SFDX or MDAPI"""
        description_elem = getattr(element, "description", None)

        # SFDX release can check this before calling us, but MDAPI can't.
        if self._should_process_object(
            version.package.namespace, sobject_name, element
        ):
            if description_elem and description_elem.text:
                description = description_elem.text
            else:
                description = ""

            self.sobjects[sobject_name].append(
                SObjectDetail(
                    version=version,
                    api_name=sobject_name,
                    label=element.label.text,
                    description=description,
                )
            )

        # For MDAPI-format elements. No-op on SFDX.
        if self._should_process_object_fields(sobject_name, element):
            for field in element.findall("fields"):
                self._process_field_element(sobject_name, field, version)
        else:
            # If this is an object type from which we shouldn't process any fields,
            # track it in omit_sobjects so we can drop any fields later if we don't have
            # the right information at time of processing.

            # Note that the owning object may be in a dependency package, so we won't find it otherwise.
            self.omit_sobjects.add(sobject_name)

    def _process_field_element(
        self,
        sobject: str,
        field: metadata_tree.MetadataElement,
        version: PackageVersion,
    ):
        """Process a field entity, which can be either a <fields> element
        in MDAPI format or a <CustomField> in SFDX"""
        # `element` may be either a `fields` element (in MDAPI)
        # or a `CustomField` (SFDX)
        # If this is a custom field, register its presence in this version
        if "__" in field.fullName.text:
            field_name = f"{version.package.namespace}{field.fullName.text}"
            # get field help text value
            help_text_elem = field.find("inlineHelpText")
            # get field description text value
            description_text_elem = field.find("description")

            field_type = field.type.text
            valid_values = ""

            length = ""

            if not field.find("formula"):
                if field_type in ["Text", "LongTextArea"]:
                    length = f" ({field.length.text})"
                elif field_type == "Number":
                    length = f" ({int(field.precision.text) - int(field.scale.text)}.{field.scale.text})"

            if field_type in ("Picklist", "MultiselectPicklist"):
                # There's two different ways of storing picklist values
                # (exclusive of Global Value Sets).
                # <picklist> is used prior to API 38.0:
                # https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta/api_meta/meta_picklist.htm
                # <valueSet> is used thereafter:
                # https://developer.salesforce.com/docs/atlas.en-us.api_meta.meta/api_meta/meta_field_types.htm#meta_type_valueset
                if field.find("valueSet") is not None:
                    # Determine if this field uses a Global Value Set.
                    value_set = field.valueSet
                    if value_set.find("valueSetName") is not None:
                        value_set_name = value_set.find("valueSetName").text
                        valid_values = f"Global Value Set {value_set_name}"
                    else:
                        valueSetDefinition = value_set.valueSetDefinition
                        labels = [
                            value.find("label").text
                            if value.find("label")
                            else value.fullName.text
                            for value in valueSetDefinition.findall("value")
                        ]

                        valid_values = "; ".join(labels)
                elif field.find("picklist") is not None:
                    picklist = field.picklist
                    names = [
                        value.fullName.text
                        for value in picklist.findall("picklistValues")
                    ]

                    valid_values = "; ".join(names)
            elif field_type == "Lookup":
                target_sobject = field.referenceTo.text
                if target_sobject.count("__") == 1:
                    target_sobject = f"{version.package.namespace}{target_sobject}"
                field_type = f"Lookup to {target_sobject}"
                # Note: polymorphic custom fields are not allowed.
            elif field_type == "MasterDetail":
                target_sobject = field.referenceTo.text
                if target_sobject.count("__") == 1:
                    target_sobject = f"{version.package.namespace}{target_sobject}"
                field_type = f"Master-Detail Relationship to {target_sobject}"
                # Note: polymorphic custom fields are not allowed.

            if help_text_elem and help_text_elem.text:
                help_text = help_text_elem.text
            else:
                help_text = ""

            if description_text_elem and description_text_elem.text:
                description = description_text_elem.text
            else:
                description = ""

            fd = FieldDetail(
                version=version,
                sobject=sobject,
                api_name=field_name,
                label=field.label.text,
                type=f"{field_type}{length}",
                help_text=help_text,
                description=description,
                valid_values=valid_values,
            )
            fully_qualified_name = f"{sobject}.{fd.api_name}"
            self.fields[fully_qualified_name].append(fd)

    @staticmethod
    def _get_version_name(package_version: PackageVersion) -> str:
        if package_version.version == PRERELEASE_SIGIL:
            return "Prerelease"
        return str(package_version.version)

    def _write_object_results(self, file_handle):
        """Write to the given handle an output CSV containing the data dictionary for sObjects."""
        writer = csv.writer(file_handle, quoting=csv.QUOTE_ALL)

        writer.writerow(
            [
                "Object Label",
                "Object API Name",
                "Object Description",
                "Version Introduced",
                "Version Deleted",
            ]
        )

        for sobject_name, versions in self.sobjects.items():
            # object_version.version.version yields the LooseVersion
            # of the package version of this object version.
            versions = sorted(
                versions,
                key=lambda object_version: object_version.version.version,
                reverse=True,
            )
            first_version = versions[-1]
            last_version = versions[0]

            # Locate the version, if any, where this object was deleted.
            package_versions = sorted(
                self.package_versions[last_version.version.package]
            )
            if last_version.version.version != package_versions[-1]:
                deleted_version = package_versions[
                    package_versions.index(last_version.version.version) + 1
                ]
            else:
                deleted_version = None

            writer.writerow(
                [
                    last_version.label,
                    sobject_name,
                    last_version.description,
                    f"{first_version.version.package.package_name} {self._get_version_name(first_version.version)}",
                    ""
                    if deleted_version is None
                    else f"{first_version.version.package.package_name} {deleted_version}",
                ]
            )

    def _write_field_results(self, file_handle):
        """Write to the given handle an output CSV containing the data dictionary for fields."""
        writer = csv.writer(file_handle, quoting=csv.QUOTE_ALL)

        writer.writerow(
            [
                "Object Label",
                "Object API Name",
                "Field Label",
                "Field API Name",
                "Type",
                "Picklist Values",
                "Help Text",
                "Field Description",
                "Version Introduced",
                "Version Picklist Values Last Changed",
                "Version Help Text Last Changed",
                "Version Deleted",
            ]
        )

        for _, field_versions in self.fields.items():
            # field_version.version.version yields the LooseVersion of the package version for this field version.
            versions = sorted(
                field_versions,
                key=lambda field_version: field_version.version.version,
                reverse=True,
            )
            first_version = versions[-1]
            last_version = versions[0]

            if last_version.sobject in self.omit_sobjects:
                continue

            # Locate the last versions where the valid values and the help text changed.
            valid_values_version = None
            for (index, version) in enumerate(versions[1:]):
                if version.valid_values != last_version.valid_values:
                    valid_values_version = versions[index]
                    break

            help_text_version = None
            for (index, version) in enumerate(versions[1:]):
                if version.help_text != last_version.help_text:
                    help_text_version = versions[index]
                    break

            # Locate the version, if any, where this field was deleted.
            package_versions = sorted(
                self.package_versions[last_version.version.package]
            )
            if last_version.version.version != package_versions[-1]:
                deleted_version = package_versions[
                    package_versions.index(last_version.version.version) + 1
                ]
            else:
                deleted_version = None

            # Find the sObject name, if possible, for this field.
            if last_version.sobject in self.sobjects:
                sobject_label = sorted(
                    self.sobjects[last_version.sobject],
                    key=lambda ver: ver.version.version,
                    reverse=True,
                )[0].label
            else:
                sobject_label = last_version.sobject

            writer.writerow(
                [
                    sobject_label,
                    last_version.sobject,
                    last_version.label,
                    last_version.api_name,
                    last_version.type,
                    last_version.valid_values,
                    last_version.help_text,
                    last_version.description,
                    f"{first_version.version.package.package_name} {self._get_version_name(first_version.version)}",
                    f"{first_version.version.package.package_name} {self._get_version_name(valid_values_version.version)}"
                    if valid_values_version
                    else "",
                    f"{first_version.version.package.package_name} {self._get_version_name(help_text_version.version)}"
                    if help_text_version
                    else "",
                    ""
                    if deleted_version is None
                    else f"{first_version.version.package.package_name} {deleted_version}",
                ]
            )

    def _write_results(self):
        """Write the stored schema details to our destination CSVs"""
        with open(self.options["object_path"], "w") as object_file:
            self._write_object_results(object_file)

        with open(self.options["field_path"], "w") as field_file:
            self._write_field_results(field_file)

    def _version_from_tag_name(
        self, tag_name: str, prefix_release: str
    ) -> LooseVersion:
        """Parse a release's tag and return a LooseVersion"""
        return LooseVersion(tag_name[len(prefix_release) :])



import logging
import os
import xml.etree.ElementTree as ET
import zipfile
from tempfile import NamedTemporaryFile
from unittest.mock import MagicMock, patch

import pytest

from cumulusci.core.config import TaskConfig
from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.salesforce_api.retrieve_profile_api import RetrieveProfileApi
from cumulusci.tasks.salesforce.retrieve_profile import RetrieveProfile


@pytest.fixture
def retrieve_profile_task(tmpdir):
    project_config = MagicMock()
    task_config = TaskConfig(
        {
            "options": {
                "profiles": ["Profile1", "Profile2"],
                "path": tmpdir,
                "strict_mode": False,
            }
        }
    )
    org_config = MagicMock()
    task = RetrieveProfile(project_config, task_config, org_config)
    task.logger = logging.getLogger(__name__)
    task.logger.setLevel(logging.DEBUG)
    return task


def test_check_existing_profiles_with_missing_profiles_and_strict_mode_enabled(tmpdir):
    project_config = MagicMock()
    task_config = TaskConfig(
        {
            "options": {
                "profiles": ["Profile1", "Profile2"],
                "path": tmpdir,
                "strict_mode": True,
            }
        }
    )
    org_config = MagicMock()
    retrieve_profile_task = RetrieveProfile(project_config, task_config, org_config)
    with patch.object(
        RetrieveProfileApi, "_retrieve_existing_profiles", return_value=["Profile1"]
    ):
        with pytest.raises(
            RuntimeError, match="Operation failed due to missing profiles"
        ):
            retrieve_profile_task._check_existing_profiles(RetrieveProfileApi)


def test_check_existing_profiles_with_no_existing_profiles(tmpdir):
    project_config = MagicMock()
    task_config = TaskConfig(
        {
            "options": {
                "profiles": ["Profile1", "Profile2"],
                "path": tmpdir,
                "strict_mode": False,
            }
        }
    )
    org_config = MagicMock()
    retrieve_profile_task = RetrieveProfile(project_config, task_config, org_config)
    with patch.object(
        RetrieveProfileApi, "_retrieve_existing_profiles", return_value=[]
    ):
        with pytest.raises(
            RuntimeError, match="None of the profiles given were found."
        ):
            retrieve_profile_task._check_existing_profiles(RetrieveProfileApi)


def test_init_options(retrieve_profile_task):
    retrieve_profile_task._init_options(retrieve_profile_task.task_config.config)
    assert retrieve_profile_task.profiles == ["Profile1", "Profile2"]
    assert retrieve_profile_task.strictMode is False


def test_init_options_raises_error_with_no_profiles():
    project_config = MagicMock()
    task_config = TaskConfig({"options": {"profiles": None}})
    org_config = MagicMock()

    with pytest.raises(ValueError) as exc_info:
        RetrieveProfile(project_config, task_config, org_config)

    assert str(exc_info.value) == "At least one profile must be specified."


def test_init_options_raises_error_with_invalid_path_directory():
    project_config = MagicMock()
    task_config = TaskConfig(
        {"options": {"profiles": ["Profile1"], "path": "/nonexistent/directory"}}
    )
    org_config = MagicMock()

    with pytest.raises(FileNotFoundError) as exc_info:
        RetrieveProfile(project_config, task_config, org_config)

    expected_message = "The extract directory '/nonexistent/directory' does not exist."
    assert str(exc_info.value) == expected_message


def test_init_options_raises_error_with_non_directory_path(tmp_path):
    tmpfile = tmp_path / "file.txt"
    tmpfile.write_text("Something")
    project_config = MagicMock()
    task_config = TaskConfig({"options": {"profiles": ["Profile1"], "path": tmpfile}})
    org_config = MagicMock()

    with pytest.raises(NotADirectoryError) as exc_info:
        RetrieveProfile(project_config, task_config, org_config)

    expected_message = f"'{tmpfile}' is not a directory."
    assert str(exc_info.value) == expected_message


def create_temp_zip_file():
    temp_zipfile = NamedTemporaryFile(delete=True)

    with zipfile.ZipFile(temp_zipfile, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.writestr("profiles/Profile1.profile", "")

    return zipfile.ZipFile(temp_zipfile, "r")


def test_save_profile_file_new(retrieve_profile_task, tmpdir):
    extract_dir = str(tmpdir)
    filename = "TestProfile.profile"
    meta_filename = "TestProfile.profile-meta.xml"
    content = "Profile content"
    expected_file_path = os.path.join(extract_dir, meta_filename)
    retrieve_profile_task.save_profile_file(extract_dir, filename, content)

    assert os.path.exists(expected_file_path)
    with open(expected_file_path, "r", encoding="utf-8") as profile_file:
        saved_content = profile_file.read()
    assert saved_content == content


def test_save_profile_file_existing_meta_xml(retrieve_profile_task, tmpdir):
    extract_dir = str(tmpdir)
    filename = "TestProfile.profile"
    meta_filename = "TestProfile.profile-meta.xml"
    content = "Profile content"
    existing_file_path = os.path.join(extract_dir, meta_filename)

    with open(existing_file_path, "w", encoding="utf-8") as existing_file:
        existing_file.write("Existing content")

    retrieve_profile_task.save_profile_file(extract_dir, filename, content)

    with open(existing_file_path, "r", encoding="utf-8") as profile_file:
        saved_content = profile_file.read()
    assert saved_content == content


def test_save_profile_file_existing(retrieve_profile_task, tmpdir):
    extract_dir = str(tmpdir)
    filename = "TestProfile.profile"
    content = "Profile content"
    existing_file_path = os.path.join(extract_dir, filename)

    with open(existing_file_path, "w", encoding="utf-8") as existing_file:
        existing_file.write("Existing content")

    retrieve_profile_task.save_profile_file(extract_dir, filename, content)

    with open(existing_file_path, "r", encoding="utf-8") as profile_file:
        saved_content = profile_file.read()
    assert saved_content == content


def test_add_flow_accesses(retrieve_profile_task):
    profile_content = "<Profile>\n" "    <some_tag>Hello</some_tag>\n" "</Profile>"
    flows = ["Flow1", "Flow2"]
    expected_content = (
        "<Profile>\n"
        "    <some_tag>Hello</some_tag>\n"
        "    <flowAccesses>\n"
        "        <enabled>true</enabled>\n"
        "        <flow>Flow1</flow>\n"
        "    </flowAccesses>\n"
        "    <flowAccesses>\n"
        "        <enabled>true</enabled>\n"
        "        <flow>Flow2</flow>\n"
        "    </flowAccesses>\n"
        "</Profile>"
    )
    modified_content = retrieve_profile_task.add_flow_accesses(profile_content, flows)
    assert modified_content == expected_content

    # Content without the </Profile> tag
    profile_content = "<Profile>\n" "    <some_tag>Hello</some_tag>\n"
    modified_content = retrieve_profile_task.add_flow_accesses(profile_content, flows)
    assert modified_content == profile_content


def test_run_task(retrieve_profile_task, tmpdir, caplog):
    retrieve_profile_task.extract_dir = tmpdir
    temp_zipfile = create_temp_zip_file()

    with patch.object(
        RetrieveProfileApi,
        "_retrieve_permissionable_entities",
        return_value=({"ApexClass": ["TestApexClass"]}, {"Profile1": ["Flow1"]}),
    ), patch.object(
        RetrieveProfileApi, "_init_task", return_value="something"
    ), patch.object(
        ApiRetrieveUnpackaged, "__call__", return_value=temp_zipfile
    ), patch.object(
        RetrieveProfileApi, "_retrieve_existing_profiles", return_value=["Profile1"]
    ):
        retrieve_profile_task._run_task()

    assert os.path.exists(tmpdir)
    profile1_path = os.path.join(tmpdir, "profiles/Profile1.profile-meta.xml")
    assert os.path.exists(profile1_path)

    log_messages = [record.message for record in caplog.records]
    assert f"Profiles Profile1 unzipped into folder '{tmpdir}'" in log_messages
    assert (
        f"Profiles Profile1, Profile2 unzipped into folder '{tmpdir}'"
        not in log_messages
    )
    assert (
        "The following profiles were not found or could not be retrieved: 'Profile2'\n"
        in log_messages
    )


def test_get_api(retrieve_profile_task):
    retrieve_profile_task.package_xml = ""
    api = retrieve_profile_task._get_api()
    assert isinstance(api, ApiRetrieveUnpackaged)


def remove_whitespace(xml_str):
    return "".join(line.strip() for line in xml_str.splitlines())


def compare_xml_strings(xml_str1, xml_str2):
    xml_str1 = remove_whitespace(xml_str1)
    xml_str2 = remove_whitespace(xml_str2)

    tree1 = ET.ElementTree(ET.fromstring(xml_str1))
    tree2 = ET.ElementTree(ET.fromstring(xml_str2))

    parsed_xml_str1 = ET.tostring(tree1.getroot(), encoding="unicode")
    parsed_xml_str2 = ET.tostring(tree2.getroot(), encoding="unicode")

    return parsed_xml_str1 == parsed_xml_str2


def test_create_package_xml(retrieve_profile_task):
    input_dict = {
        "ApexClass": ["Class1", "Class2"],
        "Profile": ["Profile1", "Profile2"],
    }
    package_xml = retrieve_profile_task._create_package_xml(input_dict, "58.0")

    expected_package_xml = """<?xml version="1.0" encoding="UTF-8"?>
    <Package xmlns="http://soap.sforce.com/2006/04/metadata">
        <types>
            <members>Class1</members>
            <members>Class2</members>
            <name>ApexClass</name>
        </types>
        <types>
            <members>Profile1</members>
            <members>Profile2</members>
            <name>Profile</name>
        </types>
        <version>58.0</version>
    </Package>"""

    assert compare_xml_strings(package_xml, expected_package_xml)


if __name__ == "__main__":
    pytest.main()



import pytest

from cumulusci.core.config import (
    BaseProjectConfig,
    OrgConfig,
    ServiceConfig,
    UniversalConfig,
)
from cumulusci.core.config.scratch_org_config import ScratchOrgConfig


@pytest.fixture()
def project_config():
    universal_config = UniversalConfig()
    project_config = BaseProjectConfig(universal_config, config={"no_yaml": True})
    project_config.config["services"] = {
        "connected_app": {"attributes": {"test": {"required": True}}},
        "github": {"attributes": {"name": {"required": True}, "password": {}}},
        "github_enterprise": {
            "attributes": {"name": {"required": True}, "password": {}}
        },
        "not_configured": {"attributes": {"foo": {"required": True}}},
        "devhub": {"attributes": {"foo": {"required": True}}},
        "marketing_cloud": {
            "class_path": "cumulusci.core.config.marketing_cloud_service_config.MarketingCloudServiceConfig",
            "attributes": {"foo": {"required": True}},
        },
    }
    project_config.project__name = "TestProject"
    return project_config


@pytest.fixture
def org_config():
    return OrgConfig({"foo": "bar"}, "test")


@pytest.fixture
def scratch_org_config():
    return ScratchOrgConfig({"foo": "bar"}, "test")


@pytest.fixture
def key():
    return "0123456789123456"


@pytest.fixture
def service_config():
    return ServiceConfig({"name": "bar@baz.biz", "password": "test123"})



"""Salesforce customizations of VCR

Hides session IDs and OrgIDs.
Generally hides all request and response headers to save space in our repo.

Records transactions if there is an org specified and does not if there is not.
"""

import re
from functools import partial
from pathlib import Path

import pytest
from vcr import cassette

from cumulusci.core.enums import StrEnum

from .pytest_sf_vcr_serializer import CompressionVCRSerializer


def simplify_body(request_or_response_body):
    decoded = request_or_response_body.decode("utf-8")
    decoded = _cleanup(decoded)

    return decoded.encode()


class RecordingMode(StrEnum):
    RECORD = "Recording"
    READ = "Reading"
    DISABLE = "Disabled"


replacements = [
    (r"/v?\d\d.0/", r"/vxx.0/"),
    (r"/00D[\w\d]{12,15}", "/00D0xORGID00000000"),
    (r'"00D[\w\d]{12,15}"', '"00D0xORGID00000000"'),
    (r".com//", r".com/"),
    (r"ersion>\d\d.0<", r"ersion>vxx.0<"),
    (r"<sessionId>.*</sessionId>", r"<sessionId>**Elided**</sessionId>"),
    # (r"001[\w\d]{12,15}", "0010xACCOUNTID0000"),
    (r"//.*.my.salesforce.com", "//orgname.my.salesforce.com"),
    (r"//.*\d+.*.salesforce.com/", "//orgname.my.salesforce.com/"),
    # (
    #     r"202\d-\d\d-\d\dT\d\d:\d\d\:\d\d.\d\d\d\+0000",
    #     "2021-01-01T01:02:01.000+0000",
    # ),
    # (r'"005[\w\d]{12,15}"', '"0050xUSERID0000000"'),
    (r'"InstanceName" : "[A-Z]{2,4}\d{1,4}",', '"InstanceName" : "CS420",'),
    # (r"<id>0.*<\/id>", "<id>0ANAPPID{}</id>"),  # replace SOAP message IDs.
    (
        r"<asyncProcessId>0.*<\/asyncProcessId>",
        "<asyncProcessId>0ANAPPID</asyncProcessId>",
    ),
    # in case we ever want to normalize dates again, here is how we would do it.
    # (r"\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\d.\d\d\dZ", "2021-01-03T01:11:11.420Z"),
    # (r"/User/005[\w\d]{12,15}", "/User/0050xUSERID0000000"),
    # (r"<lastModifiedById>005[\w\d]{12,15}", "<lastModifiedById>005USERID"),
]

replacements = [
    (re.compile(pattern), replacement) for (pattern, replacement) in replacements
]


@pytest.fixture(scope="function")
def vcr_cassette_path(vcr_cassette_dir, vcr_cassette_name):
    return Path(vcr_cassette_dir, vcr_cassette_name + ".yaml")


class VcrState:
    recording: RecordingMode = RecordingMode.READ


@pytest.fixture(scope="session")
def vcr_state():
    return VcrState()


@pytest.fixture(autouse=True)
def configure_recording_mode(
    request,
    user_requested_network_access,
    vcr_cassette_path,
    user_requested_cassette_replacement,
    vcr_state,
    monkeypatch,
):
    recording_mode = vcr_state.recording
    if (
        user_requested_network_access
        and vcr_cassette_path.exists()
        and user_requested_cassette_replacement
    ):
        vcr_cassette_path.unlink()
        recording_mode = RecordingMode.RECORD
    elif user_requested_network_access and vcr_cassette_path.exists():
        # user wants to keep existing cassette, so disable VCR usage entirely, like:
        # https://github.com/ktosiek/pytest-vcr/blob/08482cf0724697c14b63ad17752a0f13f7670add/pytest_vcr.py#L59
        recording_mode = RecordingMode.DISABLE
    elif user_requested_network_access:
        recording_mode = RecordingMode.RECORD
    else:
        # reading
        recording_mode = RecordingMode.READ

    with monkeypatch.context() as m:
        m.setattr(vcr_state, "recording", recording_mode)
        yield


def sf_before_record_request(vcr_state, http_request):
    if vcr_state.recording == RecordingMode.DISABLE:
        return None
    if http_request.body:
        http_request.body = simplify_body(http_request.body)
    http_request.uri = _cleanup(http_request.uri)

    http_request.headers = {"Request-Headers": "Elided"}

    return http_request


def sf_before_record_response(response):
    response["headers"] = {
        "Content-Type": response["headers"].get("Content-Type", "None"),
        "Others": "Elided",
    }
    if response.get("body"):
        response["body"]["string"] = simplify_body(response["body"]["string"])
    return response


def vcr_config(request, user_requested_network_access, vcr_state):
    "Fixture for configuring VCR"

    # https://vcrpy.readthedocs.io/en/latest/usage.html#record-modes
    if user_requested_network_access:
        record_mode = "new_episodes"  # should this be once?
    else:
        record_mode = "none"

    return {
        "record_mode": record_mode,
        "decode_compressed_response": True,
        "before_record_response": sf_before_record_response,
        "before_record_request": partial(sf_before_record_request, vcr_state),
        # this is redundant, but I guess its a form of
        # security in-depth
        "filter_headers": [
            "Authorization",
            "Cookie",
            "Public-Key-Pins-Report-Only",
            "Last-Modified",
        ],
    }


@pytest.fixture(scope="function")
def bind_vcr_state(request, vcr_state):
    "Give the vcr_state object access to the request"
    vcr_state.request = request
    yield
    vcr_state.request = None


def _cleanup(s: str):
    if s:
        s = str(s, "utf-8") if isinstance(s, bytes) else s
        for pattern, replacement in replacements:
            s = pattern.sub(replacement, s)
        return s


def explain_mismatch(r1, r2):
    print("CURRENT", r1, "\nTAPED", r2)
    for a, b in zip(r1, r2):
        if a != b:
            print("\nMISMATCH\n\t Current:", a, "\n!=\n\tTaped:   ", b)
            break
    return False


def salesforce_matcher(r1, r2, should_explain=False):
    summary1 = (r1.method, _cleanup(r1.uri), _cleanup(r1.body))
    summary2 = (r2.method, _cleanup(r2.uri), _cleanup(r2.body))
    # uncomment explain_mismatch if you need to debug.
    # otherwise it will generate a lot of noise, even when things
    # are working properly
    if summary1 != summary2:
        if should_explain:
            return None
            # return explain_mismatch(summary1, summary2)
        else:
            assert summary1 == summary2

    return True


@pytest.fixture(scope="session")
def salesforce_serializer(shared_vcr_cassettes):
    return CompressionVCRSerializer(shared_vcr_cassettes)


def salesforce_vcr(vcr, salesforce_serializer):
    vcr.register_matcher("Salesforce Matcher", salesforce_matcher)
    vcr.match_on = ["Salesforce Matcher"]
    vcr.register_serializer("Compression Serializer", salesforce_serializer)
    vcr.serializer = "Compression Serializer"
    return vcr


salesforce_vcr.__doc__ = __doc__

orig_contains = cassette.Cassette.__contains__


# better error handling than the built-in VCR stuff
def __contains__(self, request):
    """Return whether or not a request has been stored"""
    if orig_contains(self, request):
        return True

    # otherwise give a helpful warning
    for index, response in self._responses(request):
        if self.play_counts[index] != 0:
            raise AssertionError(
                f"SALESFORCE VCR Error: Request matched but response had already been used **** {request}"
            )

    for index, (stored_request, response) in enumerate(self.data):
        salesforce_matcher(request, stored_request, should_explain=True)

    return False


@pytest.fixture(
    scope="function",
)
def run_code_without_recording(
    request, vcr, user_requested_network_access, vcr_state, monkeypatch
):
    def really_run_code_without_recording(func):
        if user_requested_network_access:
            # Run the setup code, but don't record it
            with monkeypatch.context() as m:
                m.setattr(vcr_state, "recording", RecordingMode.DISABLE)
                return func()

    return really_run_code_without_recording


cassette.Cassette.__contains__ = __contains__



import logging
import typing as T

from simple_salesforce import Salesforce
from sqlalchemy import Column, MetaData, Table, Unicode, UniqueConstraint, and_
from sqlalchemy.engine.base import Connection
from sqlalchemy.exc import IntegrityError

from cumulusci.core.exceptions import BulkDataException
from cumulusci.tasks.bulkdata.mapping_parser import CaseInsensitiveDict, MappingStep
from cumulusci.tasks.bulkdata.query_transformers import LoadQueryExtender
from cumulusci.tasks.bulkdata.step import DataOperationType
from cumulusci.tasks.bulkdata.utils import create_table_if_needed, sf_query_to_table

logger = logging.getLogger(__name__)


def _create_empty_upsert_key_table(
    tablename, metadata, fieldnames: T.List[str]
) -> Table:
    """Create an empty table matching the upsert semantics"""
    fields = [Column(fieldname, Unicode(255)) for fieldname in fieldnames]

    # the data fields must be unique
    non_id_fields = [fieldname for fieldname in fieldnames if fieldname.lower() != "id"]
    assert len(fields) == len(non_id_fields) + 1
    fields.append(UniqueConstraint(*non_id_fields))

    return create_table_if_needed(tablename, metadata, fields)


def upsert_tablename_for_obj_and_keys(sf_object, keys):
    return f"upsert_{sf_object}_{'_'.join(keys)}"


def extract_upsert_key_data(
    sf_object: str,
    keys: tuple,
    context,
    metadata: MetaData,
    connection: Connection,
) -> Table:
    """Create a table with keys and IDs from Salesforce"""
    tablename = upsert_tablename_for_obj_and_keys(sf_object, keys)
    if "Id" not in (key.title() for key in keys):
        keys = ("Id",) + keys

    table = _create_empty_upsert_key_table(tablename, metadata, keys)

    soql_query = f"select {','.join(keys)} from {sf_object}"
    try:
        logger.info(f"Extracting records for upsert: `{soql_query}`")
        table = sf_query_to_table(
            table=table,
            sobject=sf_object,
            fields=keys,
            api_options={},
            context=context,
            query=soql_query,
            metadata=metadata,
            connection=connection,
        )
    except IntegrityError as e:
        message = str(e)
        if "UNIQUE constraint failed" not in message:  # pragma: no cover
            raise e
        raise BulkDataException(f"Duplicate values for upsert key:\n {e.params}") from e

    return table


def needs_etl_upsert(mapping: MappingStep, sf: Salesforce):
    """Is this an upsert that Salesforce cannot do natively?"""
    # is this an upsert and one that Salesforce cannot do by itself?
    keys = mapping.update_key

    if len(keys) > 1:
        return True

    describe_data = mapping.describe_data(sf)

    key = keys[0]
    is_sf_builtin_key = (
        key == "Id"
        or describe_data[key]["externalId"]
        or describe_data[key]["idLookup"]
    )
    return not is_sf_builtin_key


class AddUpsertsToQuery(LoadQueryExtender):
    @property
    def columns_to_add(self):
        if self.mapping.action == DataOperationType.ETL_UPSERT:
            # We've retrieved IDs from the org, so include them.

            # If we treat "Id" as an "external_id_name" then it's
            # allowed to be sparse.
            tablename = upsert_tablename_for_obj_and_keys(
                self.mapping.sf_object, self.mapping.update_key
            )
            table = self.metadata.tables[tablename]
            id_column = table.columns["Id"]
            return [id_column]

    @property
    def outerjoins_to_add(self):
        if self.mapping.action == DataOperationType.ETL_UPSERT:
            tablename = upsert_tablename_for_obj_and_keys(
                self.mapping.sf_object, self.mapping.update_key
            )
            table = self.metadata.tables[tablename]
            return [self._make_outerjoin_for_upsert(table)]

    def _make_outerjoin_for_upsert(
        self,
        upsert_key_table: Table,
    ):
        """Make a query that joins user-provided data to Salesforce IDs"""
        update_data_table = self.metadata.tables[self.mapping.table]

        # normalize columns
        update_table_columns = CaseInsensitiveDict(update_data_table.c.items())

        join_on_clauses = [
            (upsert_key_table.columns[column_name] == update_table_columns[column_name])
            for column_name in self.mapping.update_key
        ]

        return (upsert_key_table, and_(*join_on_clauses))



import os

import factory
import pytest

from cumulusci.tasks.bulkdata import factory_utils
from cumulusci.tasks.bulkdata.tests.dummy_data_factory import (
    ContactFactory,
    GenerateDummyData,
)
from cumulusci.tasks.bulkdata.tests.utils import _make_task
from cumulusci.utils import temporary_dir


class TestFactoryUtils:
    def test_factory(self):
        mapping_file = os.path.join(os.path.dirname(__file__), "mapping_vanilla_sf.yml")

        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "temp.db")
            dburl = "sqlite:///" + tmp_db_path
            task = _make_task(
                GenerateDummyData,
                {
                    "options": {
                        "num_records": 12,
                        "mapping": mapping_file,
                        "database_url": dburl,
                    }
                },
            )
            task()


class TestAdder:
    def test_adder(self):
        a = factory_utils.Adder(10)
        b = a(20)
        assert b == 30
        c = a(0)
        assert c == 30
        d = a(-5)
        assert d == 25
        a.reset(3)
        assert a(0) == 3


class TestFactories:
    def test_factories(self):
        class Broken(factory.alchemy.SQLAlchemyModelFactory):
            class Meta:
                model = "xyzzy"

        with pytest.raises(KeyError):
            factory_utils.Factories(None, {}, {"A": ContactFactory, "B": Broken})



import re
from typing import Callable, Dict, NamedTuple

from robot.api import ExecutionResult, ResultVisitor  # type: ignore
from robot.result.model import TestCase

UNITS = {
    "setup_time": ("s", "Setup Time"),
    "teardown_time": ("s", "Teardown Time"),
    "total_time": ("s", "Total Time"),
    "elapsed_time": ("s", "Elapsed Time"),
}


class PerfSummary(NamedTuple):
    name: str
    metrics: Dict[str, float]
    test: TestCase


def _format_metric(metric_name: str, value: float) -> str:
    "Format a metric value for output"
    extra_info = UNITS.get(metric_name)
    if extra_info:
        unit, name = extra_info
    else:
        unit, name = "", metric_name
    return f"{name}: {value}{unit} "


def _perf_formatter(perfsummary: PerfSummary) -> str:
    """Default formatter for perf summaries"""

    other_metrics = [
        _format_metric(metric, value)
        for metric, value in perfsummary.metrics.items()
        if metric != "total_time"
    ]

    result = f"{perfsummary.name} - "

    if other_metrics:
        other_metrics = ", ".join(other_metrics)
        result += other_metrics

    return result


def log_perf_summary_from_xml(
    robot_xml, logger_func: Callable, formatter_func: Callable = _perf_formatter
):
    """Log Robot performance info to a callable which accepts a string.

    Supply a formatter that takes a PerfSummary triple if the default isn't a good fit:

        f(test_name: str, metrics: Dict[str, float], test: robot.result.model.TestCase)
    """
    result = ExecutionResult(robot_xml)
    pl = _perf_logger(logger_func, formatter_func)
    next(pl)  # start the generator
    perf_summarizer = PerfSummarizer(pl.send)
    result.visit(perf_summarizer)


def _perf_logger(logger_func: Callable, formatter_func: Callable):
    """Generator that connects visitor to logger"""
    # ensure we have at least one result before printing header
    perfsummary = yield
    logger_func(" === Performance Results  === ")

    while perfsummary:
        logger_func(formatter_func(perfsummary))
        perfsummary = yield


pattern = r"\${cci_metric_(?P<metric>.*)} = (?P<value>.*)"
# used also in MetaCI
ELAPSED_TIME_PATTERN = re.compile(pattern)


class PerfSummarizer(ResultVisitor):
    """Robot ResultVisitor that looks for performance summaries"""

    def __init__(self, callable):
        self.times = []
        self.callable = callable

    def start_test(self, test):
        self.metrics = {}

    def end_message(self, message):
        # n.b. we're looking for messages like '${cci_metric_foo} = x'
        match = ELAPSED_TIME_PATTERN.match(message.message)
        if match:
            try:
                self.metrics[match["metric"]] = float(match["value"])
            except ValueError as e:  # pragma: no cover
                raise ValueError(  # -- should never happen
                    f"Cannot find number in {message.message}"
                ) from e

    def end_test(self, test):
        if self.metrics:
            self.report_test_metrics(test, self.metrics)

    def report_test_metrics(self, test, metrics):
        "Generate a perf summary and call formatter&logger to output it"
        if test.setup:
            metrics["setup_time"] = test.setup.elapsedtime / 1000
        if test.teardown:
            metrics["teardown_time"] = test.teardown.elapsedtime / 1000
        metrics["total_time"] = test.elapsedtime / 1000

        self.callable(PerfSummary(test.name, metrics, test))



import logging
import os
import warnings
from functools import lru_cache
from typing import Any, Dict, Optional

# turn on strictness.
# delete this when strictness is mandatory
STRICT_GETATTR = os.environ.get("STRICT_GETATTR")
CHECK_CONFIG_TYPES = os.environ.get("CHECK_CONFIG_TYPES")


class BaseConfig(object):
    """BaseConfig provides a common interface for nested access for all Config objects in CCI."""

    defaults = {}
    config: dict
    logger: logging.Logger

    def __init__(self, config: Optional[dict] = None, keychain=None):
        if config is None:
            self.config = {}
        else:
            types = self._all_allowed_names()
            if CHECK_CONFIG_TYPES:
                for k, v in config.items():
                    type_for_value = types.get(k)
                    if not type_for_value:
                        warnings.warn(f"{k}: {v} not declared for {type(self)}")
                    if (v is not None) and (type_for_value is not None):
                        assert isinstance(
                            v, type_for_value
                        ), f"{k}: {v} should be of type {type_for_value}, not {type(v)} for {type(self)}"
            self.config = config.copy()

        self._init_logger()
        self._load_config()

    def _init_logger(self):
        """Initializes self.logger"""
        self.logger = logging.getLogger(__name__)

    def _load_config(self):
        """Subclasses may override this method to initialize :py:attr:`~config`"""
        pass

    @classmethod
    def _allowed_names(cls) -> Dict[str, type]:
        return getattr(cls, "__annotations__", {})

    # long term plan is to get rid of this
    def __getattr__(self, name: str) -> Any:
        """Look up a property in a sub-dictionary

        Property names should be declared in each Config class with type annotations.
        """
        if not name.startswith("_"):
            first_part = name.split("__")[0]
            if first_part not in self._all_allowed_names():
                message = (
                    f"Property `{first_part}` is unknown on class `{self.__class__.__name__}`. "
                    + "Either declare it in the type declaration or use lookup() to look it up dynamically"
                )
                warnings.warn(
                    message,
                    DeprecationWarning,
                )

                assert not STRICT_GETATTR, message
        return self.lookup(name, already_called_getattr=True)

    @classmethod
    @lru_cache
    def _all_allowed_names(cls) -> Dict[str, type]:
        "Allowed names from this class and its base classes"
        allowed_names_from_all_base_classes = (
            baseclass._allowed_names()  # type: ignore
            for baseclass in cls.__mro__
            if hasattr(baseclass, "_allowed_names")
        )
        ret = {}
        for d in allowed_names_from_all_base_classes:
            ret.update(d)
        return ret

    def lookup(
        self, name: str, default: Any = None, already_called_getattr: bool = False
    ) -> Any:
        tree = name.split("__")
        if name.startswith("_"):
            raise AttributeError(f"Attribute {name} not found")
        value: Any = None
        value_found = False
        config = self.config
        if len(tree) > 1:
            # Walk through the config dictionary using __ as a delimiter
            for key in tree[:-1]:
                config = config.get(key)
                if config is None:
                    break
        if config and tree[-1] in config:
            value = config[tree[-1]]
            value_found = True

        if value_found:
            return value
        else:
            if not already_called_getattr and hasattr(self, name):
                return getattr(self, name)
            else:
                return self.defaults.get(name, default)



import os
from unittest import mock

import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce import EnsureRecordTypes
from cumulusci.utils import temporary_dir

from .util import create_task

OPPORTUNITY_METADATA = """<?xml version="1.0" encoding="utf-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <businessProcesses>
        <fullName>NPSP_Default</fullName>
        <isActive>true</isActive>
        <values>
            <fullName>Test</fullName>
            <default>false</default>
        </values>
    </businessProcesses>
    <recordTypes>
        <fullName>NPSP_Default</fullName>
        <active>true</active>
        <businessProcess>NPSP_Default</businessProcess>
        <label>NPSP Default</label>
        <description></description>
    </recordTypes>
</CustomObject>
"""

CASE_METADATA = """<?xml version="1.0" encoding="utf-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <businessProcesses>
        <fullName>NPSP_Default</fullName>
        <isActive>true</isActive>
        <values>
            <fullName>Test</fullName>
            <default>true</default>
        </values>
    </businessProcesses>
    <recordTypes>
        <fullName>NPSP_Default</fullName>
        <active>true</active>
        <businessProcess>NPSP_Default</businessProcess>
        <label>NPSP Default</label>
        <description>The first 255 characters of record_type_descirption option-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------</description>
    </recordTypes>
</CustomObject>
"""

ACCOUNT_METADATA = """<?xml version="1.0" encoding="utf-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    
    <recordTypes>
        <fullName>NPSP_Default</fullName>
        <active>true</active>
        
        <label>NPSP Default</label>
        <description>Default Account Record Type created by NPSP.</description>
    </recordTypes>
</CustomObject>
"""  # noqa: W293

PACKAGE_XML = """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>*</members>
        <name>CustomObject</name>
    </types>
    <version>45.0</version>
</Package>"""

OPPORTUNITY_DESCRIBE_NO_RTS = {
    "recordTypeInfos": [{"master": True}],
    "fields": [
        {"name": "Id"},
        {
            "name": "StageName",
            "picklistValues": [
                {"value": "Bad", "active": False},
                {"value": "Test", "active": True},
            ],
        },
    ],
}

CASE_DESCRIBE_NO_RTS = {
    "recordTypeInfos": [{"master": True}],
    "fields": [
        {"name": "Id"},
        {
            "name": "Status",
            "picklistValues": [
                {"value": "Bad", "active": False},
                {"value": "Test", "active": True},
            ],
        },
    ],
}

OPPORTUNITY_DESCRIBE_WITH_RTS = {
    "recordTypeInfos": [{"master": False}, {"master": True}],
    "fields": [
        {"name": "Id"},
        {
            "name": "StageName",
            "picklistValues": [
                {"value": "Bad", "active": False},
                {"value": "Test", "active": True},
            ],
        },
    ],
}


class TestEnsureRecordTypes:
    def test_infers_correct_business_process(self):
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "sobject": "Opportunity",
            },
        )
        task.sf = mock.Mock()
        task.sf.Opportunity = mock.Mock()
        task.sf.Opportunity.describe = mock.Mock(
            return_value=OPPORTUNITY_DESCRIBE_NO_RTS
        )

        task._infer_requirements()

        assert task.options["generate_business_process"]
        assert task.options["generate_record_type"]
        assert task.options["stage_name"] == "Test"

    def test_no_business_process_where_unneeded(self):
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "sobject": "Account",
            },
        )
        task.sf = mock.Mock()
        task.sf.Account = mock.Mock()
        task.sf.Account.describe = mock.Mock(return_value=OPPORTUNITY_DESCRIBE_NO_RTS)

        task._infer_requirements()

        assert not task.options["generate_business_process"]
        assert task.options["generate_record_type"]
        assert "stage_name" not in task.options

    def test_generates_record_type_and_business_process(self):
        # Asserts Record Type Description is optional.
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "sobject": "Opportunity",
            },
        )

        task.sf = mock.Mock()
        task.sf.Opportunity = mock.Mock()
        task.sf.Opportunity.describe = mock.Mock(
            return_value=OPPORTUNITY_DESCRIBE_NO_RTS
        )
        task._infer_requirements()

        with temporary_dir():
            task._build_package()
            with open(os.path.join("objects", "Opportunity.object"), "r") as f:
                opp_contents = f.read()
                assert OPPORTUNITY_METADATA == opp_contents
                assert OPPORTUNITY_METADATA == opp_contents
            with open(os.path.join("package.xml"), "r") as f:
                pkg_contents = f.read()
                assert PACKAGE_XML == pkg_contents

    def test_generates_record_type_and_business_process__case(self):
        # Asserts Record Type Description is added and truncated to 255 characters.
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "record_type_description": "The first 255 characters of record_type_descirption option-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------Everything past here is more than 255 characters",
                "sobject": "Case",
            },
        )

        task.sf = mock.Mock()
        task.sf.Case = mock.Mock()
        task.sf.Case.describe = mock.Mock(return_value=CASE_DESCRIBE_NO_RTS)
        task._infer_requirements()

        with temporary_dir():
            task._build_package()
            with open(os.path.join("objects", "Case.object"), "r") as f:
                opp_contents = f.read()
                assert CASE_METADATA == opp_contents
            with open(os.path.join("package.xml"), "r") as f:
                pkg_contents = f.read()
                assert PACKAGE_XML == pkg_contents

    def test_generates_record_type_only(self):
        # Asserts Record Type Description is added when the Description is less than 255 characters.
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "record_type_description": "Default Account Record Type created by NPSP.",
                "sobject": "Account",
            },
        )

        task.sf = mock.Mock()
        task.sf.Account = mock.Mock()
        task.sf.Account.describe = mock.Mock(return_value=OPPORTUNITY_DESCRIBE_NO_RTS)
        task._infer_requirements()

        with temporary_dir():
            task._build_package()
            with open(os.path.join("objects", "Account.object"), "r") as f:
                opp_contents = f.read()
                assert ACCOUNT_METADATA == opp_contents
            with open(os.path.join("package.xml"), "r") as f:
                pkg_contents = f.read()
                assert PACKAGE_XML == pkg_contents

    def test_no_action_if_existing_record_types(self):
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "sobject": "Opportunity",
            },
        )
        task.sf = mock.Mock()
        task.sf.Opportunity = mock.Mock()
        task.sf.Opportunity.describe = mock.Mock(
            return_value=OPPORTUNITY_DESCRIBE_WITH_RTS
        )

        task._infer_requirements()

        assert not task.options["generate_business_process"]
        assert not task.options["generate_record_type"]

    def test_second_rt_if_force_create(self):
        # Asserts a second Record Type is added even when RTs already exist when force_create is True
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "record_type_description": "Default Account Record Type created by NPSP.",
                "sobject": "Account",
                "force_create": True,
            },
        )

        task.sf = mock.Mock()
        task.sf.Account = mock.Mock()
        # no impact from using opp describe; it's simply representing Record Types as present
        task.sf.Account.describe = mock.Mock(return_value=OPPORTUNITY_DESCRIBE_WITH_RTS)
        task._infer_requirements()

        with temporary_dir():
            task._build_package()
            with open(os.path.join("objects", "Account.object"), "r") as f:
                obj_contents = f.read()
                assert ACCOUNT_METADATA == obj_contents
            with open(os.path.join("package.xml"), "r") as f:
                pkg_contents = f.read()
                assert PACKAGE_XML == pkg_contents

    def test_executes_deployment(self):
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "sobject": "Opportunity",
            },
        )

        task.sf = mock.Mock()
        task.sf.Opportunity = mock.Mock()
        task.sf.Opportunity.describe = mock.Mock(
            return_value=OPPORTUNITY_DESCRIBE_NO_RTS
        )
        task._deploy = mock.Mock()
        task._run_task()

        task._deploy.assert_called_once()
        task._deploy.return_value.assert_called_once()

    def test_no_deployment_if_unneeded(self):
        task = create_task(
            EnsureRecordTypes,
            {
                "record_type_developer_name": "NPSP_Default",
                "record_type_label": "NPSP Default",
                "sobject": "Opportunity",
            },
        )

        task.sf = mock.Mock()
        task.sf.Opportunity = mock.Mock()
        task.sf.Opportunity.describe = mock.Mock(
            return_value=OPPORTUNITY_DESCRIBE_WITH_RTS
        )
        task._deploy = mock.Mock()
        task._run_task()

        task._deploy.assert_not_called()

    def test_init_options(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                EnsureRecordTypes,
                {
                    "record_type_developer_name": "%#$(*%(%))",
                    "record_type_label": "NPSP Default",
                    "sobject": "Opportunity",
                },
            )

        with pytest.raises(TaskOptionsError):
            create_task(
                EnsureRecordTypes,
                {
                    "record_type_developer_name": "NPSP_Default",
                    "record_type_label": "NPSP Default",
                    "sobject": "Test__c",
                },
            )
